
/*************************************************************************/
/* ECG_AUTO.HPP                                                      E13 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

namespace ecg {

/*************************************************************************/
/* AUTO::Compile()                                                       */
/*                                                                       */
/*************************************************************************/
template <class T, class CONFIG>
void AUTO<T, CONFIG>::Compile(T & context)
{
	// SET FIRST STATE
	int state = 0;
	{
		STATE st = { 0, 0, 0, 0 };
		lalrStack.push(st);
	}

	// PARSE FIRST TOKEN
	int token = context.NextToken();

	enum {
		NORMAL,
		ERROR_OCCURED,
		CLEANUP,
	} compileState = NORMAL;

	ACTION	action;
	bool	wasCleanUp = false;
	int		lastCleanUpState = -1;
	//bool	errorShifted = false;
	//int		lastErrorShiftReduceState = -1;

	#define ECG_CALL_PROC(PROC)									\
		if (PROC != NULL) {										\
			assert(!context.compilerState.is(CST_FATAL_ERROR));	\
			(context.*(PROC))();								\
		}

	#define ECG_CALL_CHECK(PROC)								\
		if (PROC != NULL) {										\
			if (context.compilerState.is(CST_FATAL_ERROR)) {	\
				return ;										\
			}													\
			/* check whether user wants to cleanup */			\
			if (context.compilerState.is(CST_CLEANUP)) {		\
				compileState = CLEANUP;							\
				break;											\
			}													\
			wasCleanUp = false;									\
		}

	for (;;) {
		assert((size_t)state < STATE_COUNT);

		switch (compileState) {
			case NORMAL:
				assert(wasCleanUp == context.compilerState.is(CST_CLEANUP));

				// look up action in table
				if (token == TOKEN_EOF) {
					action = actionTableEof[state];
				} else if (token == TOKEN_ERROR) {
					action = actionTableError[state];
					//if (action.type == SHIFT) {
					//	errorShifted = true;
					//}
				} else {
					assert(token >= 0 && token < TERM_COUNT);
					action = (*actionTable)[state][token];
				}

//CHECK_ACTION_TYPE:
				if (action.type == ERROR) {
					//// find any look-ahead symbol that allows reduce
					//if (errorShifted && state != lastErrorShiftReduceState) {
					//	bool found = false;
					//	ACTION tmpAction;
					//	for (int i=0; i<TERM_COUNT; ++i) {
					//		tmpAction = (*actionTable)[state][i];
					//		if (tmpAction.type == REDUCE) {
					//			found = true;
					//			break;
					//		}
					//	}

					//	if (found) {
					//		lastErrorShiftReduceState = state;

					//		errorShifted = false;
					//		action = tmpAction;
					//		goto CHECK_ACTION_TYPE;
					//	}
					//}

					if (!wasCleanUp) {
						compileState = ERROR_OCCURED;
						break;
					}

					//lastErrorShiftReduceState = -1;

					assert(context.compilerState.is(CST_CLEANUP));

					// check if there's nothing to skip
					if (token == TOKEN_EOF) {
						context.compilerState.set(CST_FATAL_ERROR);

						// call error message (unexpected end of file)
						if (actionTableError[state].type == ERROR) {
							context.Error();
						}
						return ;
					}

					// skip bad tokens
					token = context.NextToken();
					break;
				}

				assert(action.type == SHIFT || action.type == REDUCE);
				switch (action.type) {
					case SHIFT:
						assert(action.data < STATE_COUNT);
						// store backup data with new state
						state = action.data;
						{
							STATE st = {
								state,
								(int)context.sstack.size(),
								(int)context.vstack.size(),
								(int)context.pstack.size(),
							};
							lalrStack.push(st);
						}

						// call semantic procedure for terminal
						if (token != TOKEN_EOF && token != TOKEN_ERROR) {
							assert(token >= 0 && token < TERM_COUNT);

							ECG_CALL_PROC(procTableTerm[token]);
						}

						if (token == TOKEN_ERROR) {
							// TOKEN_ERROR is virtual token (don't skip it)
							token = context.Token();
						} else {
							token = context.NextToken();
						}
						break;

					case REDUCE:
						{
#ifdef _DEBUG
							int prevState = state;
#endif
							assert(action.data >= 0 && action.data < RULE_COUNT);
							PRODUCTION prod = procTableProduction[action.data];

							// skip this reduce states and pop new state
							for (int i=0; i<prod.stateCount; ++i) {
								lalrStack.pop();
							}
							state = lalrStack.top().state;
							assert(state >= 0 && state < STATE_COUNT);

							state = (*productionTable)[state][prod.nonterm];

							// check for happy state ;)
							if (state == TOKEN_EOF) {
								assert(lalrStack.size() == 1);
								assert(context.sstack.empty());
								assert(context.vstack.empty());
								assert(context.pstack.empty());
								lalrStack.c.clear();

								// OK!!!!
								return ;
							}

							// call semantic procedure for the production
							ECG_CALL_PROC(prod.proc);

							// store backup data with new state
							{
								STATE st = {
									state,
									(int)context.sstack.size(),
									(int)context.vstack.size(),
									(int)context.pstack.size(),
								};
								lalrStack.push(st);
							}

							ECG_CALL_CHECK(prod.proc);

							// call semantic procedure for nonterminal
							assert(prod.nonterm >= 0 && prod.nonterm < NONTERM_COUNT);
							ECG_CALL_PROC(procTableNonterm[prod.nonterm]);
							assert(wasCleanUp || !context.compilerState.is(CST_CLEANUP));
						}
						break;
				}
				break;

			case ERROR_OCCURED:
				context.compilerState.set(CST_ERROR_OCCURED);
				lastCleanUpState = -1; // clear last cleanup state

				// override token
				token = TOKEN_ERROR;

				// try to reduce known error-handle rule
				if (actionTableError[state].type != ERROR) {
					compileState = NORMAL;
					//errorShifted = false;
					break;
				}

				// call default error proc
				context.compilerState.set(CST_FATAL_ERROR);
				context.Error();

				// user wants to clean up
				if (context.compilerState.is(CST_CLEANUP)) {
					context.compilerState.clear(CST_FATAL_ERROR);
					compileState = CLEANUP;
					break;
				}

				if (context.compilerState.is(CST_FATAL_ERROR)) {
					return ;
				}

				// all ok
				compileState = NORMAL;
				//errorShifted = false;
				break;

			case CLEANUP:
				context.compilerState.set(CST_ERROR_OCCURED);

				// override token
				token = TOKEN_ERROR;

				// search for clean up rule
				assert(!lalrStack.empty());
				if (actionTableError[state].type == ERROR) {
					do {
						// clean state stack
						lalrStack.pop();

						if (lalrStack.empty()) {
							// unable to find clean up rule, this is the fatal error
							context.compilerState.set(CST_FATAL_ERROR);

							// don't call error proc because compiler already knows about it
							return ;
						}

						// pick up next state
						state = lalrStack.top().state;
					} while (actionTableError[state].type == ERROR);

					// roll back stacks after cleaning
					{
						#define ECG_ROLLBACK_STACK(STACK, SIZE)		\
							assert((int)(STACK).size() >= SIZE);	\
							while ((int)(STACK).size() > SIZE) {	\
								(STACK).pop();						\
							}

						STATE & st = lalrStack.top();

						ECG_ROLLBACK_STACK(context.sstack, st.sstackSize);
						ECG_ROLLBACK_STACK(context.vstack, st.vstackSize);
						ECG_ROLLBACK_STACK(context.pstack, st.pstackSize);
					}
				}

				wasCleanUp   = true;
				//errorShifted = false;
				compileState = NORMAL;

				if (state == lastCleanUpState) {
					token = context.NextToken();
					break;
				}

				lastCleanUpState = state;
				break;
		}
	}
}

} // namespace ecg