
#ifndef ECG_OS_H
#define ECG_OS_H

#pragma once

namespace ecg {

typedef struct {} * HANDLE;

enum FILE_OPEN {
	FO_READ,
	FO_WRITE,
	FO_APPEND,
};

enum FILE_SEEK {
	FS_BEGIN,
	FS_CURRENT,
	FS_END,
};

HANDLE			FileOpen		(const char * name, FILE_OPEN type);
int				FileRead		(HANDLE file, void * dst, int length);
int				FileSeek		(HANDLE file, int position, FILE_SEEK method = FS_BEGIN);
void			FileClose		(HANDLE file);

void				FolderCreate	(const char * path);

int				Timer			(void); // precision 1 ms

typedef __int64 int64;
int64			Timer2			(void);	// precision 1e-3 ms (Timer2() / Timer2Freq())
int64			Timer2Freq		(void);

struct TIMER {
	TIMER(bool start = false)
		: timer(start ? Timer() : 0)
		, counter()
	{}

	void		Start			(void)
	{ timer = Timer(); }

	int			Snapshot		(void)
	{ return counter = Timer() - timer; }

	int 		Counter			(void) const
	{ return counter; }

	bool		Elapsed			(int delay, bool restart = false)
	{ if (Snapshot() < delay) return false; if (restart) Start(); return true; }

private:
	int		timer;
	int		counter;
};

struct TIMER2 { // double precision
	TIMER2(bool start = false)
		: timer(start ? Timer2() : 0)
		, counter()
	{}

	void		Start			(void)
	{ timer = Timer2(); }

	float		Snapshot		(void)
	{ return counter = (Timer2() - timer) / (float)Timer2Freq(); }

	float 		Counter			(void) const
	{ return counter; }

	bool		Elapsed			(int delay, bool restart = false)
	{ if (Snapshot() < delay) return false; if (restart) Start(); return true; }

private:
	int64		timer;
	float		counter;
};

template <class LEX>
struct OS_IO_POLICY {
	OS_IO_POLICY(void)
		: file(), remains(), buf(), bufCur(), error()
	{}

	~OS_IO_POLICY(void)
	{ delete [] buf; }

	enum { BUFFER_SIZE = 4096 };

	bool Init(const char * name, std::ostream & error)
	{
		file = FileOpen(name, FO_READ);
		if (file == NULL) {
			return false;
		}

		if (buf == NULL) {
			buf = new char[BUFFER_SIZE];
		}

		this->error		= &error;
		return true;
	}

	inline bool GetChar(typename LEX::CHAR_T & ch)
	{
		if (remains == 0) {
			remains = FileRead(file, buf, BUFFER_SIZE);
			if (remains == 0) {
				FileClose(file);
				file = NULL;
				return false;
			}
			bufCur = buf;
		}

		ch = *bufCur++;
		--remains;
		return true;
	}

protected:
	std::ostream &			GetError			(void) const { return *error; }

private:
	HANDLE					file;
	char *					buf;
	char *					bufCur;
	int						remains;
	std::ostream *			error;
};

} // namespace ecg

#endif // ECG_OS_H