
/*************************************************************************/
/* ECG_GENERATOR.CPP                                          06.11.2005 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

#include "stdafx.h"
#include "ecg_generator.h"

#define for if (false) ; else for

using namespace std;
using namespace ecg;

/*************************************************************************/
/* GENERATOR::LoadGrammatics()                                           */
/*                                                                       */
/*************************************************************************/
bool GENERATOR::LoadGrammatics(istream & input, ostream & error)
{
	loader.Init(*this, input, error);

	loader.Compile();
	if (loader.HasError()) {
		return false;
	}

	//    
	for (NONTERMINAL_ARRAY::iterator nt = nonterminal.begin(); nt != nonterminal.end(); ++nt) {
		if (nt->rules.empty()) {
			error << "ERROR: Nonterminal \"" << nt->name << "\" has no rules" << endl;
			error << endl;
			return false;
		}
	}

	return true;
}

/*************************************************************************/
/* GENERATOR::AddNonterminal()                                           */
/*                                                                       */
/*************************************************************************/
void GENERATOR::AddNonterminal(const string & name)
{
	if (nonterminalNumber.find(name) == nonterminalNumber.end()) {
		NONTERMINAL nt;
		nt.name = name;
		nt.id = (int)nonterminal.size();
		nonterminal.push_back(nt);
		nonterminalNumber.insert(make_pair(nt.name, nt.id));
		assert(nonterminal.size() == nonterminalNumber.size());
	}
}

/*************************************************************************/
/* GENERATOR::AddTerminal()                                              */
/*                                                                       */
/*************************************************************************/
void GENERATOR::AddTerminal(const string & name)
{
	if (terminalNumber.find(name) == terminalNumber.end()) {
		TERMINAL term;
		term.id = (int)terminal.size();
		term.name = name;
		term.alias = termPrefix + TransformNameToAlias(name);
		terminal.push_back(term);
		terminalNumber.insert(make_pair(name, term.id));
		assert(terminal.size() == terminalNumber.size());
	}
}

/*************************************************************************/
/* struct IS_VALID_NAME_CHAR                                             */
/*                                                                       */
/*************************************************************************/
struct IS_BAD_NAME_CHAR {
	bool operator () (char ch) const
	{
		char_traits<char>::int_type ich = char_traits<char>::to_int_type(ch);
		return !isalpha(ich) && !isdigit(ich) && strchr("_", ich) == NULL;
	}
};

/*************************************************************************/
/* GENERATOR::TransformNameToAlias()                                     */
/*                                                                       */
/*************************************************************************/
string GENERATOR::TransformNameToAlias(const string & name) const
{
	STRING_STRING_MAP::const_iterator iter = aliasTA.find(name);
	if (iter != aliasTA.end()) {
		aliasSet.insert(iter->second);
		return iter->second;
	}

	string result = name[0] == '$' ? name.c_str() + 1 : name;
	assert(strlen(&result[0]) == result.size());
	strupr(&result[0]);

	if (find_if(result.begin(), result.end(), IS_BAD_NAME_CHAR()) != result.end()
	||  aliasSet.find(result) != aliasSet.end()) {
		char buf[32];
		result = string("NAME_") + itoa(++aliasCounter, buf, 10);
	}
	aliasSet.insert(result);
	return result;
}

/*************************************************************************/
/* GENERATOR::CalcFirst()                                                */
/*                                                                       */
/*************************************************************************/
void GENERATOR::CalcFirst(void)
{
	for (NONTERMINAL_ARRAY::iterator nonterm = nonterminal.begin(); nonterm != nonterminal.end(); ++nonterm) {
		nonterm->first.clear();
	}

	bool flag = true;
	while (flag) {
		flag = false;

		for (NONTERMINAL_ARRAY::iterator nt = nonterminal.begin(); nt != nonterminal.end(); ++nt) {
			//   
			for (RULE_ARRAY::iterator rule = nt->rules.begin(); rule != nt->rules.end(); ++rule) {
				RULE::ITEM_ARRAY::iterator item;
				//    
				for (item = rule->items.begin(); item != rule->items.end(); ++item) {
					//   ,     
					if (item->isTerm) {
						if (nt->first.find(item->id) == nt->first.end()) {
							flag = true;
							nt->first.insert(item->id);
						}
						break;
					}
					bool haseps = false;
					//   ,     
					for (INT_SET::iterator fir = nonterminal[item->id].first.begin(); fir != nonterminal[item->id].first.end(); ++fir) {
						if (*fir == TERMINAL_EPS) {
							haseps = true;
						} else {
							if (nt->first.find(*fir) == nt->first.end()) {
								flag = true;
								nt->first.insert(*fir);
							}
						}
					}
					//     $eps,  
					if (!haseps) {
						break;
					}
				}
				//      $eps,   $eps
				if (item == rule->items.end()) {
					if (nt->first.find(TERMINAL_EPS) == nt->first.end()) {
						flag = true;
						nt->first.insert(TERMINAL_EPS);
					}
				}
			} // rule
		} // nt
	} // while (flag)
}

/*************************************************************************/
/* GENERATOR::PrintFirst()                                               */
/*                                                                       */
/*************************************************************************/
void GENERATOR::PrintFirst(ostream & output)
{
	output << endl << "#first" << endl << endl;
	size_t maxName = 0;
	for (NONTERMINAL_ARRAY::iterator nt = nonterminal.begin(); nt != nonterminal.end(); ++nt) {
		maxName = max(maxName, nt->name.size());
	}

	for (NONTERMINAL_ARRAY::iterator nt = nonterminal.begin(); nt != nonterminal.end(); ++nt) {
		output << nt->name << string(maxName - nt->name.size(), ' ') << " :";
		for (INT_SET::iterator fir = nt->first.begin(); fir != nt->first.end(); ++fir) {
			if (TERMINAL_EPS == *fir) {
				output << " $eps";
			} else {
				output << " " << terminal[*fir].name;
			}
		}
		output << endl;
	}
	output << endl;
}

/*************************************************************************/
/* GENERATOR::MakeTerminalStep()                                         */
/*                                                                       */
/*************************************************************************/
void GENERATOR::MakeTerminalStep(const LALR_ITEM_SET & state, int termId, int & minindex, int stateIndex)
{
	LALR_ITEM item;
	LALR_ITEM_SET nextState;

	//   
	for (LALR_ITEM_SET::const_iterator it = state.begin(); it != state.end(); ++it) {
		assert(nonterminal[it->nontermId].id == it->nontermId);
		assert(nonterminal[it->nontermId].rules.size() > (size_t)it->ruleNumber);
		RULE & rule = nonterminal[it->nontermId].rules[it->ruleNumber];
		assert((size_t)it->position <= rule.items.size());
		if (it->position < (int)rule.items.size() &&
			rule.items[it->position].isTerm &&
			rule.items[it->position].id == termId)
		{
			item = *it;
			++item.position;
			nextState.insert(item);
		}
	}

	if (!nextState.empty()) {
		bool isNewState	= true;
		int oldState	= -1;

		//         
		for (size_t j=0; j<lalrStates.size(); ++j) {
			if (nextState == lalrStates[j]) {
				assert(nextState.size() == lalrStates[j].size());
				// ,   lookahead    
				bool eq = true;
				for (LALR_ITEM_SET::iterator it1 = nextState.begin(), it2 = lalrStates[j].begin(); it1 != nextState.end(); ++it1, ++it2) {
					if (!IsSubset(it1->lookahead, it2->lookahead)) {
						eq = false;
						break;
					}
				}

				//   ,   stepback,     lookahead
				if (!eq) {
					minindex = min(minindex, (int)j);
					for (LALR_ITEM_SET::iterator it1 = nextState.begin(), it2 = lalrStates[j].begin(); it1 != nextState.end(); ++it1, ++it2) {
						for (INT_SET::iterator la = it1->lookahead.begin(); la != it1->lookahead.end(); ++la) {
							it2->lookahead.insert(*la);
						}
					}
				}

				//     
				isNewState	= false;
				oldState	= (int)j;
				break;
			}
		}

		assert(oldState != -1 || isNewState);
		//  -slr   ,  ,  
		LALR_TABLE_ITEM tableItem;
		tableItem.isShift	= true;
		tableItem.state		= oldState;

		if (isNewState) {
			tableItem.state = (int)lalrStates.size();
			lalrStates.push_back(nextState);
			lalrTable.push_back(LALR_TABLE_LINE());
			assert(lalrTable.size() == lalrStates.size());
		}

		lalrTable[stateIndex].insert(make_pair(RULE::ITEM(true, termId), tableItem));
	} // !next_state.empty()
}

/*************************************************************************/
/* GENERATOR::MakeNonterminalStep()                                      */
/*                                                                       */
/*************************************************************************/
void GENERATOR::MakeNonterminalStep(const LALR_ITEM_SET & state, int nontermId, int & minindex, int stateIndex)
{
	LALR_ITEM item;
	LALR_ITEM_SET nextState;

	//   
	for (LALR_ITEM_SET::const_iterator it = state.begin(); it != state.end(); ++it) {
		assert(nonterminal[it->nontermId].id == it->nontermId);
		RULE & rule = nonterminal[it->nontermId].rules[it->ruleNumber];
		assert((size_t)it->position <= rule.items.size());
		if (it->position < (int)rule.items.size() &&
			!rule.items[it->position].isTerm &&
			rule.items[it->position].id == nontermId)
		{
			item = *it;
			++item.position;
			nextState.insert(item);
		}
	}

	if (!nextState.empty()) {
		bool isNewState	= true;
		int oldState	= -1;

		//      
		for (size_t j=0; j<lalrStates.size(); ++j) {
			if (nextState == lalrStates[j]) {
				// ,   lookahead    
				bool eq = true;
				for (LALR_ITEM_SET::iterator it1 = nextState.begin(), it2 = lalrStates[j].begin(); it1 != nextState.end(); ++it1, ++it2) {
					if (!IsSubset(it1->lookahead, it2->lookahead)) {
						eq = false;
						break;
					}
				}

				//   ,   stepback
				if (!eq) {
					minindex = min(minindex, (int)j);
					for (LALR_ITEM_SET::iterator it1 = nextState.begin(), it2 = lalrStates[j].begin(); it1 != nextState.end(); ++it1, ++it2) {
						for (INT_SET::iterator la = it1->lookahead.begin(); la != it1->lookahead.end(); ++la) {
							it2->lookahead.insert(*la);
						}
					}
				}

				isNewState = false;
				oldState = (int)j;
				break;
			}
		}

		assert(oldState != -1 || isNewState);
		//  -slr   ,  ,  
		LALR_TABLE_ITEM tableItem;
		tableItem.isShift = true;
		tableItem.state = oldState;

		if (isNewState) {
			tableItem.state = (int)lalrStates.size();
			lalrStates.push_back(nextState);
			lalrTable.push_back(LALR_TABLE_LINE());
			assert(lalrStates.size() == lalrTable.size());
		}

		lalrTable[stateIndex].insert(make_pair(RULE::ITEM(false, nontermId), tableItem));
	} // !next_state.empty()
}

/*************************************************************************/
/* GENERATOR::CheckUnreachableNonterminals()                             */
/*                                                                       */
/*************************************************************************/
void GENERATOR::CheckUnreachableNonterminals(const string & startNonterm, ostream & error, bool & hasError)
{
	map<string, bool> reachable;
	queue<string> nontermQueue;

	nontermQueue.push(startNonterm);
	reachable[startNonterm] = true;
	while (!nontermQueue.empty()) {
		string nontermName = nontermQueue.front();
		nontermQueue.pop();

		assert(nonterminalNumber.find(nontermName) != nonterminalNumber.end());
		int nontermId = nonterminalNumber[nontermName];

		assert((size_t)nontermId < nonterminal.size());
		assert(nonterminal[nontermId].name == nontermName);
		const NONTERMINAL & nonterm = nonterminal[nontermId];
		for (RULE_ARRAY::const_iterator rule = nonterm.rules.begin(); rule != nonterm.rules.end(); ++rule) {
			for (RULE::ITEM_ARRAY::const_iterator item = rule->items.begin(); item != rule->items.end(); ++item) {
				if (!item->isTerm) {
					assert((size_t)item->id < nonterminal.size());
					assert(nonterminal[item->id].id == item->id);
					string reachableNontermName = nonterminal[item->id].name;
					if (!reachable[reachableNontermName]) {
						reachable[reachableNontermName] = true;
						nontermQueue.push(reachableNontermName);
					}
				}
			}
		}
	}

	for (NONTERMINAL_ARRAY::const_iterator nonterm = nonterminal.begin(); nonterm != nonterminal.end(); ++nonterm) {
		if (!reachable[nonterm->name]) {
			error << "ERROR: Unreachable nonterminal \"" << nonterm->name << "\"" << endl;
			hasError = true;
		}
	}
}

/*************************************************************************/
/* GENERATOR::ProcessItemReduce()                                        */
/*                                                                       */
/*************************************************************************/
void GENERATOR::ProcessItemReduce(const LALR_ITEM & item, int stateId, const LALR_ITEM_SET & state, ostream & error, bool & hasError)
{
	assert(item.nontermId == nonterminal[item.nontermId].id);
	assert((size_t)item.ruleNumber < nonterminal[item.nontermId].rules.size());
	RULE & rule = nonterminal[item.nontermId].rules[item.ruleNumber];

	assert((size_t)item.position <= rule.items.size());
	//         
	if (item.position == rule.items.size()) {
		//   lookahead-
		for (INT_SET::const_iterator la = item.lookahead.begin(); la != item.lookahead.end(); ++la) {
			assert((int)*la < (int)terminal.size());
			LALR_TABLE_LINE::iterator iter = lalrTable[stateId].find(RULE::ITEM(true, *la));

			LALR_TABLE_ITEM lalrItem;
			lalrItem.isShift	= false;
			lalrItem.furl		= item;

			// without conflict
			if (iter == lalrTable[stateId].end()) {
				//         
				lalrTable[stateId].insert(make_pair(RULE::ITEM(true, *la), lalrItem));
				continue;
			}

			// find prefer for this situation
			PREFER prefer = FindPrefer(item, iter->second, *la, error);

			if (iter->second.isShift) {
				assert(prefer != PREFER_OLD_REDUCE);

				// shift-reduce conflict
				if (prefer == PREFER_SHIFT) {
					// do nothing, just ignore this production
					continue;
				} else if (prefer == PREFER_REDUCE) {
					// override previous action
					lalrTable[stateId][RULE::ITEM(true, *la)] = lalrItem;
					continue;
				}

				// mark error occurred
				hasError = true;

				// unsolved shift-reduce conflict
				error << "ERROR: Shift-reduce conflict on state " << (int)stateId << ":" << endl;

				// print item
				error << "[";
				PrintLalrItem(error, item);
				error << "]" << endl;

				// can't be shift-reduce conflict on EOF
				assert(*la != TOKEN_EOF);

				// print terminal
				if (*la == TOKEN_ERROR) {
					error << "> Terminal     = $error" << endl;
				} else {
					assert((size_t)*la < terminal.size());
					if (terminal[*la].name[0] == '$') {
						error << "> Terminal     = " << terminal[*la].name << endl;
					} else {
						error << "> Terminal     = \"" << terminal[*la].name << "\"" << endl;
					}
				}

				// print state
				error << "> State        = " << (int)stateId << endl;
				for (LALR_ITEM_SET::const_iterator stateItem = state.begin(); stateItem != state.end(); ++stateItem) {
					error << "[";
					PrintLalrItem(error, *stateItem);
					error << "]" << endl;
				}

				// next state
				error << "> Next state   = " << iter->second.state << endl;

				// make closing
				LALR_ITEM_SET dbgState;
				assert((size_t)iter->second.state < lalrStates.size());
				Closing(lalrStates[iter->second.state], dbgState);

				for (LALR_ITEM_SET::const_iterator dbgStateItem = dbgState.begin(); dbgStateItem != dbgState.end(); ++dbgStateItem) {
					error << "[";
					PrintLalrItem(error, *dbgStateItem);
					error << "]" << endl;
				}
				error << endl;

				continue;
			}

			assert(prefer != PREFER_SHIFT);

			// reduce-reduce conflict
			if (prefer == PREFER_OLD_REDUCE) {
				// old production, just skip this
				continue;
			} else if (prefer == PREFER_REDUCE) {
				// override old production
				lalrTable[stateId][RULE::ITEM(true, *la)] = lalrItem;
				continue;
			}

			// check same production
			if (item.nontermId == iter->second.furl.nontermId
			&&  item.ruleNumber == iter->second.furl.ruleNumber) {
				continue;
			}

			// mark error
			hasError = true;

			// unsolved reduce-reduce conflict
			error << "ERROR: Reduce-reduce conflict on state " << (int)stateId << ":" << endl;
			error << "[";
			PrintLalrItem(error, item);
			error << "]" << endl;

			// print terminal
			if (TOKEN_EOF == *la) {
				error << "> Terminal     = $eof" << endl;
			} else if (TOKEN_ERROR == *la) {
				error << "> Terminal     = $error" << endl;
			} else {
				assert((size_t)*la < terminal.size());
				if (terminal[*la].name[0] == '$') {
					error << "> Terminal     = " << terminal[*la].name << endl;
				} else {
					error << "> Terminal     = \"" << terminal[*la].name << "\"" << endl;
				}
			}

			// print state
			error << "> State        = " << (int)stateId << endl;
			for (LALR_ITEM_SET::const_iterator stateItem = state.begin(); stateItem != state.end(); ++stateItem) {
				error << "[";
				PrintLalrItem(error, *stateItem);
				error << "]" << endl;
			}

			// print production number
			error << "> (Nonterminal, Rule) = (" << iter->second.furl.nontermId << ", " << iter->second.furl.ruleNumber << ")" << endl;
			assert(iter->second.furl.nontermId == nonterminal[iter->second.furl.nontermId].id);
			assert((size_t)iter->second.furl.ruleNumber < nonterminal[iter->second.furl.nontermId].rules.size());

			// print rule
			const NONTERMINAL & nonterm = nonterminal[iter->second.furl.nontermId];
			error << "[";
			PrintRule(error, nonterm.id, iter->second.furl.ruleNumber);
			error << "]" << endl;

			error << endl;
		} // la != item->lookahead.end()
	} // item.position == rule.items.size()
}

/*************************************************************************/
/* GENERATOR::CalcStatesAndTable()                                       */
/*                                                                       */
/*************************************************************************/
bool GENERATOR::CalcStatesAndTable(const std::string & startNonterm, std::ostream & error)
{
	bool hasError = false;

	STRING_INT_MAP::iterator oldStart = nonterminalNumber.find(startNonterm);
	if (oldStart == nonterminalNumber.end()) {
		error << "Nonterminal with name '" << startNonterm << "' doesn't exist" << endl;
		return false;
	}

	//   
	CheckUnreachableNonterminals(startNonterm, error, hasError);

	//    S' -> S;
	NONTERMINAL start;
	//  
	{
		char buf[32];
		int counter = 0;
		do {
			start.name = startNonterm + "_" + itoa(++counter, buf, 10);
		} while (nonterminalNumber.find(start.name) != nonterminalNumber.end());
	}
	start.id = (int)nonterminal.size();
	start.first = nonterminal[oldStart->second].first;
	RULE rule;
	rule.items.push_back(RULE::ITEM(false, oldStart->second));
	start.rules.push_back(rule);

	nonterminalNumber.insert(make_pair(start.name, start.id));
	nonterminal.push_back(start);

	//   
	lalrStates.clear();
	lalrTable.clear();

	LALR_ITEM_SET	state;

	// add first table line
	{
		LALR_ITEM		item;
		item.nontermId	= start.id;
		item.ruleNumber	= 0;
		item.position	= 0;
		item.lookahead.insert(TOKEN_EOF);
		state.insert(item);
		lalrStates.push_back(state);
		lalrTable.push_back(LALR_TABLE_LINE());
	}

	//  
	for (size_t i=0; i<lalrStates.size(); ++i) {
		Closing(lalrStates[i], state);

		int minindex = (int)i+1;

		// Make steps for all terminals
		for (TERMINAL_ARRAY::iterator term = terminal.begin(); term != terminal.end(); ++term) {
			MakeTerminalStep(state, term->id, minindex, (int)i);
		}

		// Make steps for $error terminal
		MakeTerminalStep(state, TOKEN_ERROR, minindex, (int)i);

		//     
		for (NONTERMINAL_ARRAY::iterator nonterm = nonterminal.begin(); nonterm != nonterminal.end(); ++nonterm) {
			MakeNonterminalStep(state, nonterm->id, minindex, (int)i);
		}

		i = minindex-1;
	}

	//   
	{
		LALR_TABLE_ITEM lalrItem;
		lalrItem.isShift = true;
		lalrItem.state = TOKEN_EOF;
		lalrTable[0][RULE::ITEM(false, start.id)] = lalrItem;
	}

	//  
	//   
	for (size_t i=0; i<lalrStates.size(); ++i) {
		Closing(lalrStates[i], state);

		//   
		for (LALR_ITEM_SET::const_iterator item = state.begin(); item != state.end(); ++item) {
			ProcessItemReduce(*item, (int)i, state, error, hasError);
		}
	}

	return !hasError;
}

/*************************************************************************/
/* GENERATOR::IsPrefer()                                                 */
/*                                                                       */
/*************************************************************************/
bool GENERATOR::IsPrefer(RULE_INDEX index, int termId, PREFER_MAP & prefer)
{
	PREFER_MAP::const_iterator iter = prefer.find(index);
	return iter != prefer.end() &&
		(iter->second.empty() || iter->second.find(termId) != iter->second.end());
}

/*************************************************************************/
/* GENERATOR::FindPriority()                                             */
/*                                                                       */
/*************************************************************************/
PRIORITY GENERATOR::FindPriority(RULE_INDEX index)
{
	// check for rule priority
	if (priorityRule.find(index) != priorityRule.end()) {
		return priorityRule[index];
	}

	PRIORITY res;
	assert(res.priority == -1);

	NONTERMINAL & nt	= nonterminal[index.nontermId];
	RULE & rule			= nt.rules[index.ruleNumber];

	for (RULE::ITEM_ARRAY::const_iterator item = rule.items.begin(); item != rule.items.end(); ++item) {
		// skip nonterminals
		if (!item->isTerm) {
			continue;
		}

		// find nonterminal priority
		if (priorityOp.find(item->id) != priorityOp.end()) {
			// check for multiple priority
			if (res.priority != -1) {
				return PRIORITY();
			}

			res = priorityOp[item->id];
		}
	}

	return res;
}

/*************************************************************************/
/* GENERATOR::FindPrefer()                                               */
/*                                                                       */
/*************************************************************************/
GENERATOR::PREFER GENERATOR::FindPrefer(RULE_INDEX index, LALR_TABLE_ITEM oldItem, int termId, ostream & error)
{
	// prefer shift for new rule
	bool preferShift = IsPrefer(index, termId, this->preferShift);

	// prefer reduce for new rule
	bool preferReduce = IsPrefer(index, termId, this->preferReduce);

	if (preferShift && preferReduce) {
		error << "WARNING: prefer shift and prefer reduce simultaneously" << endl;
		error << "[";
		PrintRule(error, index.nontermId, index.ruleNumber);
		error << "]" << endl;

		return PREFER_NONE;
	}

	// priority for new rule
	PRIORITY prior = FindPriority(index);

	if (oldItem.isShift) {
		// explicit shift/reduce attrib has greater priority
		if (preferShift && !preferReduce) {
			return PREFER_SHIFT;
		} else if (preferReduce && !preferShift) {
			return PREFER_REDUCE;
		}

		TERM_PRIORITY_MAP::const_iterator opIter = priorityOp.find(termId);

		if (prior.priority == -1 || opIter == priorityOp.end()) {
			return PREFER_NONE;
		}

		PRIORITY op = opIter->second;

		// e + (e * e)
		if (prior.priority < op.priority) {
			return PREFER_SHIFT;
		}

		// (e * e) + e
		if (prior.priority > op.priority) {
			return PREFER_REDUCE;
		}

		// priority equals
		assert(prior.left == op.left);

		// (e + e) + e
		if (prior.left) {
			return PREFER_REDUCE;
		} else {
			return PREFER_SHIFT;
		}
	}

	// prefer shift for old rule
	bool preferShift2 = IsPrefer(oldItem.furl, termId, this->preferShift);

	// prefer reduce for old rule
	bool preferReduce2 = IsPrefer(oldItem.furl, termId, this->preferReduce);

	// priority for old rule
	PRIORITY oldPrior = FindPriority(oldItem.furl);

	if (preferReduce && preferReduce2) {
		error << "WARNING: prefer reduce simultaneously for both rules" << endl;
		error << "[";
		PrintRule(error, index.nontermId, index.ruleNumber);
		error << "]" << endl;
		error << "[";
		PrintRule(error, oldItem.furl.nontermId, oldItem.furl.ruleNumber);
		error << "]" << endl;

		return PREFER_NONE;
	}

	if (preferShift2 && preferReduce2) {
		error << "WARNING: prefer shift and prefer reduce simultaneously" << endl;
		error << "[";
		PrintRule(error, oldItem.furl.nontermId, oldItem.furl.ruleNumber);
		error << "]" << endl;

		return PREFER_NONE;
	}

	if (preferReduce && !preferReduce2) {
		// new rule
		return PREFER_REDUCE;
	} else if (preferReduce2 && !preferReduce) {
		// old rule
		return PREFER_OLD_REDUCE;
	}

	if (prior.priority != -1 && oldPrior.priority != -1) {
		if (prior.priority > oldPrior.priority) {
			return PREFER_REDUCE;
		}
		if (prior.priority < oldPrior.priority) {
			return PREFER_OLD_REDUCE;
		}
	}

	return PREFER_NONE;
}

/*************************************************************************/
/* GENERATOR::Closing()                                                  */
/*                                                                       */
/*************************************************************************/
void GENERATOR::Closing(const LALR_ITEM_SET & state, LALR_ITEM_SET & resultState) const
{
	resultState.clear();

	//  ,  
	queue<LALR_ITEM> qu;

	//    
	for (LALR_ITEM_SET::const_iterator item = state.begin(); item != state.end(); ++item) {
		qu.push(*item);
	}

	//     
	while (!qu.empty()) {
		LALR_ITEM item = qu.front();
		qu.pop();

		//      lookahead- ,   
		LALR_ITEM_SET::iterator iter = resultState.find(item);
		if (iter == resultState.end() || !IsSubset(item.lookahead, iter->lookahead)) {
			if (iter == resultState.end()) {
				iter = resultState.insert(item).first;
			} else {
				for (INT_SET::iterator la = item.lookahead.begin(); la != item.lookahead.end(); ++la) {
					iter->lookahead.insert(*la);
				}
			}

			assert(item.nontermId == nonterminal[item.nontermId].id);
			assert((size_t)item.ruleNumber < nonterminal[item.nontermId].rules.size());
			//     ,    
			const RULE & rule = nonterminal[item.nontermId].rules[item.ruleNumber];
			assert((size_t)item.position <= rule.items.size());
			if (item.position < (int)rule.items.size() && !rule.items[item.position].isTerm) {
				int id = rule.items[item.position].id; //  
				LALR_ITEM newItem;

				// A->a.Bb, a
				// B->.c, FIRST(ba)
				//  lookahead  newItem
				bool eps = true;
				for (size_t k = item.position + 1; k<rule.items.size(); ++k) {
					const RULE::ITEM & ruleItem = rule.items[k];

					eps = false;
					if (ruleItem.isTerm) {
						assert(ruleItem.id == terminal[ruleItem.id].id);
						newItem.lookahead.insert(ruleItem.id);
						break;
					} else {
						assert(nonterminal[ruleItem.id].id == ruleItem.id);
						const NONTERMINAL & nonterm = nonterminal[ruleItem.id];
						for (INT_SET::const_iterator fir = nonterm.first.begin(); fir != nonterm.first.end(); ++fir) {
							if (TERMINAL_EPS == *fir) {
								eps = true;
							} else {
								newItem.lookahead.insert(*fir);
							}
						}
						if (!eps) {
							break;
						}
					}
				}
				//      ,   first  $eps,
				//  lookahead-  
				if (eps) {
					for (INT_SET::iterator la = item.lookahead.begin(); la != item.lookahead.end(); ++la) {
						newItem.lookahead.insert(*la);
					}
				}

				//      ,     
				for (size_t i=0; i<nonterminal[id].rules.size(); ++i) {
					newItem.nontermId = id;
					newItem.ruleNumber = (int)i;
					newItem.position = 0;

					//    
					qu.push(newItem);
				} // i < nonterminal[id].rules.size()
			} // item.position < rule.items.size() && !rule.items[item.position].bterm
		} // iter == result.end() || !IsSubset(item.lookahead, iter->lookahead)
	} // !qu.empty()
}

/*************************************************************************/
/* GENERATOR::PrintRuleItem()                                            */
/*                                                                       */
/*************************************************************************/
void GENERATOR::PrintLalrItem(ostream & output, const LALR_ITEM & item, bool printEoln /*= false*/) const
{
	assert(nonterminal[item.nontermId].id == item.nontermId);
	output << nonterminal[item.nontermId].name << " ->";
	assert(nonterminal[item.nontermId].rules.size() > (size_t)item.ruleNumber);
	const RULE & rule = nonterminal[item.nontermId].rules[item.ruleNumber];

	for (size_t i=0; i < 1 + rule.items.size(); ++i) {
		if (item.position == i) {
			output << " |";
			if (i == rule.items.size()) {
				output << " ";
			}
		}
		if (i < rule.items.size()) {
			output << " ";
			if (rule.items[i].isTerm) {
				if (rule.items[i].id == TOKEN_ERROR) {
					output << "$error";
					continue;
				}

				assert((size_t)rule.items[i].id < terminal.size());
				assert(terminal[rule.items[i].id].name.size() > 0);
				if (terminal[rule.items[i].id].name[0] == '$') {
					output << terminal[rule.items[i].id].name;
				} else {
					output << "\"" << terminal[rule.items[i].id].name << "\"";
				}
			} else {
				output << nonterminal[rule.items[i].id].name;
			}
		}
	}
	output << "; (";
	for (INT_SET::const_iterator la = item.lookahead.begin(); la != item.lookahead.end(); ++la) {
		if (la != item.lookahead.begin()) {
			output << " ";
		}
		if (TOKEN_EOF == *la) {
			output << "$eof";
		} else {
			if (*la == TOKEN_ERROR) {
				output << TERM_ERROR_STR;
				continue;
			}

			assert((size_t)*la < terminal.size());
			assert(terminal[*la].name.size() > 0);
			if (terminal[*la].name[0] == '$') {
				output << terminal[*la].name;
			} else {
				output << "\"" << terminal[*la].name << "\"";
			}
		}
	}
	output << ")";
	if (printEoln) {
		output << endl;
	}
}

/*************************************************************************/
/* GENERATOR::PrintGrammatics()                                          */
/*                                                                       */
/*************************************************************************/
void GENERATOR::PrintGrammatics(std::ostream & output) const
{
	output << endl << "#terminals" << endl << endl;
	for (int i=0; (size_t)i < terminal.size(); ++i) {
		const TERMINAL & term = terminal[i];

		output << setw(2) << i << ") ";

		assert(term.name.size() > 0);
		if (term.name[0] == '$') {
			output << term.name << endl;
		} else {
			output << "\"" << term.name << "\"" << endl;
		}
	}

	output << endl << "#nonterminals" << endl << endl;
	for (int i=0; (size_t)i < nonterminal.size(); ++i) {
		const NONTERMINAL & nonterm = nonterminal[i];

		output << setw(2) << i << ") ";

		assert(nonterm.name.size() > 0);
		output << nonterm.name << endl;
	}

	int maxlen = 0;
	for (NONTERMINAL_ARRAY::const_iterator nonterm = nonterminal.begin(); nonterm != nonterminal.end(); ++nonterm) {
		maxlen = max(maxlen, (int)nonterm->name.size());
	}
	output << endl << "#rules" << endl << endl;
	for (int nontermId = 0; (size_t)nontermId < nonterminal.size(); ++nontermId) {
		const NONTERMINAL & nonterm = nonterminal[nontermId];
		for (int ruleId = 0; (size_t)ruleId < nonterm.rules.size(); ++ruleId) {
			PrintRule(output, nonterm.name + string(maxlen - nonterm.name.size(), ' '), nonterm.rules[ruleId].items, true);
		}
	}

	output << endl;
}

/*************************************************************************/
/* GENERATOR::PrintRule()                                                */
/*                                                                       */
/*************************************************************************/
void GENERATOR::PrintRule(ostream & output, const string & name, const RULE::ITEM_ARRAY & ruleItems, bool printEoln /* = false */) const
{
	output << name << " ->";
	for (RULE::ITEM_ARRAY::const_iterator item = ruleItems.begin(); item != ruleItems.end(); ++item) {
		if (item->isTerm) {
			if (item->id == TOKEN_ERROR) {
				output << " " << TERM_ERROR_STR;
				continue;
			}

			assert((size_t)item->id < terminal.size());
			assert(terminal[item->id].name.size() > 0);
			if (terminal[item->id].name[0] == '$') {
				output << " " << terminal[item->id].name;
			} else {
				output << " \"" << terminal[item->id].name << "\"";
			}
		} else {
			output << " " << nonterminal[item->id].name;
		}
	}
	if (ruleItems.empty()) {
		output << " ";
	}
	output << ";";
	if (printEoln) {
		output << endl;
	}
}

/*************************************************************************/
/* GENERATOR::PrintLalrStates()                                          */
/*                                                                       */
/*************************************************************************/
void GENERATOR::PrintLalrStates(std::ostream & output) const
{
	output << endl;
	for (size_t i=0; i<lalrStates.size(); ++i) {
		output << setw(3) << (int)i << ")";
		LALR_ITEM_SET state;
		Closing(lalrStates[i], state);

		for (LALR_ITEM_SET::const_iterator item = state.begin(); item != state.end(); ++item) {
			if (item != state.begin()) {
				output << string(4, ' ');
			}
			output << " [";
			PrintLalrItem(output, *item);
			output << "]" << endl;
		}
		output << endl;
	}
	output << endl;
}

/*************************************************************************/
/* GENERATOR::PrintLalrTable()                                           */
/*                                                                       */
/*************************************************************************/
void GENERATOR::PrintLalrTable(std::ostream & output) const
{
	typedef vector<int> INT_ARRAY;
	INT_ARRAY nontermWidth;
	INT_ARRAY termWidth;

	nontermWidth.resize(nonterminal.size());
	for (size_t i=0; i<nontermWidth.size(); ++i) {
		nontermWidth[i] = max(1 + (int)nonterminal[i].name.size(), 8);
	}

	termWidth.resize(terminal.size());
	for (size_t i=0; i<termWidth.size(); ++i) {
		termWidth[i] = max(1 + (int)terminal[i].name.size() + 2 * (terminal[i].name[0] != '$'), 8);
	}

	int eofWidth = max(1 + (int)strlen(TOKEN_ERROR_STR), 8);

	// 
	output << endl << "     ";
	for (size_t i=0; i<nontermWidth.size(); ++i) {
		output	<< string(nontermWidth[i] - nonterminal[i].name.size(), ' ')
				<< nonterminal[i].name;
	}

	output << " |";
	for (size_t i=0; i<termWidth.size(); ++i) {
		assert(terminal[i].name.size() > 0);
		if (terminal[i].name[0] == '$') {
			output	<< string(termWidth[i] - terminal[i].name.size(), ' ')
					<< terminal[i].name;
		} else {
			output	<< string(termWidth[i] - terminal[i].name.size() - 2, ' ')
					<< "\"" << terminal[i].name << "\"";
		}
	}
	output		<< string(eofWidth - 4, ' ') << "$eof";
	output		<< string(eofWidth - 6, ' ') << "$error";
	output		<< endl;

	//  
	for (size_t i=0; i<lalrStates.size(); ++i) {
		// 
		output << setw(3) << (int)i << ") ";

		for (NONTERMINAL_ARRAY::const_iterator nonterm = nonterminal.begin(); nonterm != nonterminal.end(); ++nonterm) {
			LALR_TABLE_LINE::const_iterator action = lalrTable[i].find(RULE::ITEM(false, nonterm->id));
			if (action == lalrTable[i].end()) {
				output << string(nontermWidth[nonterm->id], ' ');
			}else {
				output << setw(nontermWidth[nonterm->id]) << action->second.state;
			}
		}

		// 
		output << " |";
		for (TERMINAL_ARRAY::const_iterator term = terminal.begin(); term != terminal.end(); ++term) {
			LALR_TABLE_LINE::const_iterator action = lalrTable[i].find(RULE::ITEM(true, term->id));
			if (action == lalrTable[i].end()) {
				output << string(termWidth[term->id], ' ');
				continue;
			}
			if (action->second.isShift) {
				output << setw(termWidth[term->id]) << action->second.state;
			} else {
				int counter = 3;
				char buf[32];
				counter += (int)strlen(itoa(action->second.furl.nontermId, buf, 10));
				counter += (int)strlen(itoa(action->second.furl.ruleNumber, buf, 10));
				output << string(termWidth[term->id] - counter, ' ');
				output << "(" << action->second.furl.nontermId << "," << action->second.furl.ruleNumber << ")";
			}
		}

		//  
		LALR_TABLE_LINE::const_iterator action = lalrTable[i].find(RULE::ITEM(true, TOKEN_EOF));
		if (action == lalrTable[i].end()) {
			output << string(eofWidth, ' ');
		} else {
			if (action->second.isShift) {
				output << string(eofWidth - 1 - (action->second.state > 9), ' ') << action->second.state;
			} else {
				int counter = 3;
				char buf[32];
				counter += (int)strlen(itoa(action->second.furl.nontermId, buf, 10));
				counter += (int)strlen(itoa(action->second.furl.ruleNumber, buf, 10));
				output << string(eofWidth-counter, ' ');
				output << "(" << action->second.furl.nontermId << "," << action->second.furl.ruleNumber << ")";
			}
		}

		// Error token
		action = lalrTable[i].find(RULE::ITEM(true, TOKEN_ERROR));
		if (action == lalrTable[i].end()) {
			output << string(eofWidth, ' ');
		} else {
			if (action->second.isShift) {
				output << string(eofWidth - 1 - (action->second.state > 9), ' ') << action->second.state;
			} else {
				int counter = 3;
				char buf[32];
				counter += (int)strlen(itoa(action->second.furl.nontermId, buf, 10));
				counter += (int)strlen(itoa(action->second.furl.ruleNumber, buf, 10));
				output << string(eofWidth-counter, ' ');
				output << "(" << action->second.furl.nontermId << "," << action->second.furl.ruleNumber << ")";
			}
		}

		output << endl;
	}
	output << endl;
}

/*************************************************************************/
/* GENERATOR::Reset()                                                    */
/*                                                                       */
/*************************************************************************/
void GENERATOR::Reset(void)
{
	startNonterm.clear();

	terminalNumber.clear();
	terminal.clear();
	nonterminalNumber.clear();
	nonterminal.clear();
	lalrStates.clear();
	lalrTable.clear();
	preferShift.clear();

	name			= "ECG_COMPILER";
	termPrefix		= "TOKEN_T_";
	nontermPrefix	= "TOKEN_NT_";
	aliasTA.clear();
	aliasAT.clear();
	aliasCounter	= 0;

	processMsgOnly	= false;
	enableLex		= false;
	enableMessages	= false;
}

/*************************************************************************/
/* GENERATOR::Trim()								                     */
/*                                                                       */
/*************************************************************************/
string GENERATOR::Trim(const string & str)
{
	int indexLeft = 0;
	while (indexLeft < (int)str.size() && isspace(char_traits<char>::to_int_type(str[indexLeft]))) {
		++indexLeft;
	}

	int indexRight = (int)str.size() - 1;
	while (indexRight >= 0 && isspace(char_traits<char>::to_int_type(str[indexRight]))) {
		--indexRight;
	}

	if (indexLeft > indexRight && indexRight >= 0) {
		return "";
	}

	return str.substr(indexLeft, indexRight - indexLeft + 1);
}

/*************************************************************************/
/* GENERATOR::PrintStringWithPrefix()                                    */
/*                                                                       */
/*************************************************************************/
void GENERATOR::PrintStringWithPrefix(ostream & output, const string & prefix, const string & text)
{
	output << prefix;
	for (size_t index = 0; index < text.size(); ++index) {
		output << text[index];
		if (text[index] == '\n') {
			output << prefix;
		}
	}
}

/*************************************************************************/
/* GENERATOR::MakeAnonymousNontermName()                                 */
/*                                                                       */
/*************************************************************************/
string GENERATOR::MakeAnonymousNontermName(void) const
{
	char buf[32];
	int counter = 0;
	string name;
	do {
		name = string("anonymous_") + itoa(counter, buf, 10);
		++counter;
	} while (nonterminalNumber.find(name) != nonterminalNumber.end());

	return name;
}

/*************************************************************************/
/* GENERATOR::RegisterAutoFunction()                                     */
/*                                                                       */
/*************************************************************************/
string GENERATOR::RegisterAutoFunction(const AUTO_FUNCTION & af)
{
	P_AUTO_FUNCTION paf = new AUTO_FUNCTION(af);

	// first find same function
	AUTOFUNCTION_REV_MAP::const_iterator arIter = autoFuncsRev.find(paf);
	if (arIter != autoFuncsRev.end()) {
		return arIter->second;
	}

	char buf[32];
	int counter = 0;
	string self;

	// prefix with specific name for constructors
	string prefix = (af.isConstructor ? CONSTRUCTOR_METHOD : af.name) + "_Impl";
	do {
		self = prefix + itoa(counter, buf, 10);
		++counter;
	} while (autoFuncs.find(self) != autoFuncs.end());

	assert(!af.result.name.empty());
	autoFuncs.insert(make_pair(self, paf));
	autoFuncsRev.insert(make_pair(paf, self));

	assert(
		!af.isConstructor ||
		af.result.name == ecg::VOID_TYPE ||
		af.result.name == ecg::STRING_TYPE ||
		!af.result.isValue
	);

	return self;
}

/*************************************************************************/
/* GENERATOR::GetPrefer()                                                */
/*                                                                       */
/*************************************************************************/
bool GENERATOR::GetPrefer(int nontermId, int ruleId, INT_SET & specification, bool isPreferShift) const
{
	RULE_INDEX rule;
	rule.nontermId	= nontermId;
	rule.ruleNumber = ruleId;

	const PREFER_MAP & prefer = isPreferShift ? preferShift : preferReduce;

	PREFER_MAP::const_iterator pref = prefer.find(rule);

	if (pref == prefer.end()) {
		return false;
	}

	specification = pref->second;
	return true;
}

/*************************************************************************/
/* GENERATOR::RegisterPrefer()                                           */
/*                                                                       */
/*************************************************************************/
void GENERATOR::RegisterPrefer(int nontermId, int ruleId, const INT_SET & specification, bool isPreferShift)
{
	RULE_INDEX rule;
	rule.nontermId	= nontermId;
	rule.ruleNumber = ruleId;

	PREFER_MAP & prefer = isPreferShift ? preferShift : preferReduce;

	assert(prefer.find(rule) == prefer.end());
	prefer.insert(make_pair(rule, specification));
}

/*************************************************************************/
/* GENERATOR::FindRule()                                                 */
/*                                                                       */
/*************************************************************************/
int GENERATOR::FindRule(const RULE_ARRAY & rules, const RULE & toFind)
{
	for (int i = 0; (size_t)i < rules.size(); ++i) {
		// check size
		if (rules[i].items.size() != toFind.items.size()) {
			continue;
		}

		// check items
		bool equals = true;
		for (int j = 0; (size_t)j < toFind.items.size(); ++j) {
			if (rules[i].items[j].id != toFind.items[j].id
			||  rules[i].items[j].isTerm != toFind.items[j].isTerm) {
				equals = false;
				break;
			}
		}

		// if found
		if (equals) {
			return i;
		}
	}

	return -1;
}

/*************************************************************************/
/* AUTO_FUNCTION::Compare()                                              */
/*                                                                       */
/*************************************************************************/
int AUTO_FUNCTION::Compare(const std::string & a, const std::string & b)
{
	return strcmp(a.c_str(), b.c_str());
}

/*************************************************************************/
/* AUTO_FUNCTION::Compare()                                              */
/*                                                                       */
/*************************************************************************/
int AUTO_FUNCTION::Compare(const PARAM & a, const PARAM & b)
{
	if (int cmp = Compare(a.name, b.name)) {
		return cmp;
	}

	if (a.isValue && !b.isValue) {
		return -1;
	}
	if (!a.isValue && b.isValue) {
		return 1;
	}
	return 0;
}

/*************************************************************************/
/* AUTO_FUNCTION::CompareArrays()                                        */
/*                                                                       */
/*************************************************************************/
template <class T>
int AUTO_FUNCTION::CompareArrays(const std::vector<T> & a, const std::vector<T> & b)
{
	std::vector<T>::const_iterator iterA = a.begin(), iterB = b.begin();
	for (; iterA != a.end() && iterB != b.end(); ++iterA, ++iterB) {
		if (int cmp = Compare(*iterA, *iterB)) {
			return cmp;
		}
	}

	if (iterA == a.end()) {
		if (iterB == b.end()) {
			return 0;
		}
		return 1;
	}
	return -1;
}

/*************************************************************************/
/* operator < ()                                                         */
/*                                                                       */
/*************************************************************************/
bool AUTO_FUNCTION::operator < (const AUTO_FUNCTION & af) const
{
	if (isConstructor < af.isConstructor) {
		return true;
	} else if (isConstructor > af.isConstructor) {
		return false;
	}

	if (int cmp = Compare(name, af.name)) {
		return cmp < 0;
	}
	if (int cmp = Compare(result, af.result)) {
		return cmp < 0;
	}
	if (int cmp = CompareArrays(constParams, af.constParams)) {
		return cmp < 0;
	}
	if (int cmp = CompareArrays(params, af.params)) {
		return cmp < 0;
	}

	// equals
	return false;
}

/*************************************************************************/
/* GENERATOR::RegisterOpPrior()                                          */
/*                                                                       */
/*************************************************************************/
bool GENERATOR::RegisterOpPriority(int termId, bool left, int priority)
{
	if (priorityOp.find(termId) != priorityOp.end()) {
		return false;
	}

	priorityOp[termId] = PRIORITY(left, priority);

	return true;
}

/*************************************************************************/
/* GENERATOR::RegisterRulePriority()                                     */
/*                                                                       */
/*************************************************************************/
bool GENERATOR::RegisterRulePriority(RULE_INDEX rule, bool left, int priority)
{
	if (priorityRule.find(rule) != priorityRule.end()) {
		return false;
	}

	priorityRule[rule] = PRIORITY(left, priority);

	return true;
}

/*************************************************************************/
/* GENERATOR::CalcTable()                                                */
/*                                                                       */
/*************************************************************************/
bool GENERATOR::CalcTable(std::ostream & error)
{
	CalcFirst();
	return CalcStatesAndTable(startNonterm, error);
}

/*************************************************************************/
/* GENERATOR::GenerateLex()                                              */
/*                                                                       */
/*************************************************************************/
void GENERATOR::GenerateLex(std::ostream & header, std::ostream & code, std::ostream & hlp)
{
	string name = GetName();
	assert(!name.empty());
	if (islower(char_traits<char>::to_int_type(name[name.size() - 1]))) {
		name += "Lex";
	} else {
		name += "_LEX";
	}
	loader.lex.Generate(name, nameImpl, header, code, hlp, this);
}

/*************************************************************************/
/* GENERATOR::SetInfo()                                                  */
/*                                                                       */
/*************************************************************************/
void GENERATOR::SetInfo(const std::string & headerName, const std::string & codeName, const std::string & implName)
{
	this->headerName	= headerName;
	this->codeName		= codeName;
	this->implName		= implName;
	
	if (implName.empty()) {
		this->implName = nameImpl;
		strlwr(&this->implName[0]);
	}
}

//
// End-of-file ECG_GENERATOR.CPP
//
