
/*************************************************************************/
/* ECG_GENERATOR.H                                            05.11.2005 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

#ifndef ECG_GENERATOR_H
#define ECG_GENERATOR_H

#include "ecg_loader_impl.h"

#ifndef NULL
#define NULL			0
#endif

#ifndef TOKEN_EOF
#define TOKEN_EOF		-1
#endif

namespace ecg {

/*************************************************************************/
/* ecg typedefs                                                          */
/*                                                                       */
/*************************************************************************/
typedef std::set<int> INT_SET;
typedef std::vector<std::string> STRING_ARRAY;
typedef std::map<std::string, int> STRING_INT_MAP;
typedef std::map<std::string, std::string> STRING_STRING_MAP;
typedef std::set<std::string> STRING_SET;

/*************************************************************************/
/* struct TERMINAL                                                       */
/*                                                                       */
/*************************************************************************/
struct TERMINAL {
						TERMINAL			(void) : id(-1) {}

	int					id;					// id 
	std::string			name;				//  
	std::string			code;				// ,     
	std::string			alias;				// alias for this terminal that is using in code generation
};
typedef std::vector<TERMINAL> TERMINAL_ARRAY;

/*************************************************************************/
/* struct RULE                                                           */
/*                                                                       */
/*************************************************************************/
struct RULE {
	struct ITEM {
						ITEM				(void) : isTerm(false), id(-1) {}
						ITEM				(bool isTerm, int id) : isTerm(isTerm), id(id) {}

		bool			isTerm;				//    
		int				id;					// ID /

		bool			operator <			(const ITEM & ri) const
		{ return isTerm < ri.isTerm || isTerm == ri.isTerm && id < ri.id; }
	};
	typedef std::vector<ITEM> ITEM_ARRAY;

	ITEM_ARRAY			items;				//  / 
	std::string			code;				// ,   
};
typedef std::vector<RULE> RULE_ARRAY;

/*************************************************************************/
/* struct NONTERMINAL                                                    */
/*                                                                       */
/*************************************************************************/
struct NONTERMINAL {
						NONTERMINAL			(void) : id(-1) {}

	int					id;					// ID 
	std::string			name;				//  
	RULE_ARRAY			rules;				//    
	INT_SET				first;				// FIRST(name)

	std::string			code;				// code that calls on mentioned nonterminal's rule
};
typedef std::vector<NONTERMINAL> NONTERMINAL_ARRAY;

/*************************************************************************/
/* struct RULE_INDEX                                                     */
/*                                                                       */
/*************************************************************************/
struct RULE_INDEX {
	int					nontermId;			// ID 
	int					ruleNumber;			//    

	bool				operator <			(const RULE_INDEX & ri) const
	{ return nontermId < ri.nontermId || nontermId == ri.nontermId && ruleNumber < ri.ruleNumber; }

	bool				operator ==			(const RULE_INDEX & ri) const
	{ return nontermId == ri.nontermId && ruleNumber == ri.ruleNumber; }
};

/*************************************************************************/
/* struct AUTO_FUNCTION                                                  */
/*                                                                       */
/*************************************************************************/
struct AUTO_FUNCTION {
	bool				operator <			(const AUTO_FUNCTION & af) const;

	struct PARAM {
		bool			isValue;
		std::string		name;
	};

	bool				isConstructor;
	bool				isMarker;
	std::string			name;
	PARAM				result;
	STRING_ARRAY		constParams;

	typedef std::vector<PARAM> PARAM_ARRAY;

	PARAM_ARRAY			params;

	static inline int	Compare				(const std::string & a, const std::string & b);
	static inline int	Compare				(const PARAM & a, const PARAM & b);

	template <class T>
	static int			CompareArrays		(const std::vector<T> & a, const std::vector<T> & b);

	AUTO_FUNCTION		(void) : isConstructor(false), isMarker(false) {}
};
typedef P<AUTO_FUNCTION> P_AUTO_FUNCTION;

/*************************************************************************/
/* struct PRIORITY                                                       */
/*                                                                       */
/*************************************************************************/
struct PRIORITY {
	bool				left;
	int					priority;

	PRIORITY			(void) : left(false), priority(-1) {}
	PRIORITY			(bool left, int priority) : left(left), priority(priority) {}
};

const int TERMINAL_EPS							= TOKEN_EOF;			//      

static const char * TOKEN_EOF_STR				= "TOKEN_EOF";
static const char * TOKEN_ERROR_STR				= "TOKEN_ERROR";
static const char * TERM_ERROR_STR				= "$error";

static const char * STRING_TYPE					= "string";				// string result type (for now single builded-in typed)
static const char * VOID_TYPE					= "void";				// void result type (for now single builded-in typed)

static const char * PUSH_STRING_METHOD			= "intPushString";		// method name for pushing builded-in string type
static const char * PUSH_EMPTY_STRING_METHOD	= "intPushEmptyString"; // method name for pushing builded-in string type
static const char * INIT_PTR_ARRAY_METHOD_T		= "intCreatePtrArray";	// pointer array creation proc
static const char * APPEND_PTR_ARRAY_METHOD_T	= "intAppendPtrArray";		
static const char * INIT_VAL_ARRAY_METHOD_T		= "intCreateValArray";	// value array creation proc
static const char * APPEND_VAL_ARRAY_METHOD_T	= "intAppendValArray";		
static const char * INIT_STRING_ARRAY_METHOD	= "intCreateStringArray";
static const char * APPEND_STRING_ARRAY_METHOD  = "intAppendStringArray";
static const char * CONSTRUCTOR_METHOD			= "intConstructor";

static const char *	COMMENT_LINE				= "$cline";
static const char *	COMMENT_BLOCK_BEGIN			= "$cbegin";
static const char *	COMMENT_BLOCK_END			= "$cend";
static const char * COMMENT_PRESET				= "comment";
static const char * ALL_PRESET					= "all";

/*************************************************************************/
/* class GENERATOR                                                       */
/*                                                                       */
/*************************************************************************/
class GENERATOR {
	friend class ecg::LOADER;
	friend class ecg::LEX;
public:
						GENERATOR			(void) { Reset(); }

	bool				LoadGrammatics		(std::istream & input, std::ostream & error);
	bool				CalcTable			(std::ostream & error);
	void				Generate			(std::ostream & header, std::ostream & code, std::ostream & hlp);
	void				GenerateLex			(std::ostream & header, std::ostream & code, std::ostream & hlp);
	void				GenerateMsg			(std::ostream & header, std::ostream & code);

	void				Reset				(void);
	void				SetProcessMsgOnly	(bool	processMsgOnly = true) { this->processMsgOnly = processMsgOnly; }

	void				SetInfo				(const std::string & headerName, const std::string & codeName, const std::string & implName);

	std::string			GetName				(void) const { return nameImpl; }
	std::string			GetImplName			(void) const { return name; }
	bool				IsLexEnabled		(void) const { return enableLex; }
	bool				IsMessagesEnabled	(void) const { return enableMessages; }

	void				PrintGrammatics		(std::ostream & output) const;

	void				PrintFirst			(std::ostream & output);
	void				PrintLalrStates		(std::ostream &	output) const;
	void				PrintLalrTable		(std::ostream &	output) const;

	static std::string	Trim				(const std::string & str);

private:
	void				CalcFirst			(void);
	bool				CalcStatesAndTable	(const std::string & startNonterm, std::ostream & error);

private:
	// TYPES
	struct LALR_ITEM : RULE_INDEX {
		int				position;			//   
		INT_SET			lookahead;			//  look-ahead 

		bool			operator <			(const LALR_ITEM & item) const
		{ return RULE_INDEX::operator < (item) || RULE_INDEX::operator == (item) && position < item.position; }

		bool			operator ==			(const LALR_ITEM & item) const
		{ return RULE_INDEX::operator == (item) && position == item.position; }

		LALR_ITEM		(void) : position(), RULE_INDEX() {}
	};
	typedef std::set<LALR_ITEM> LALR_ITEM_SET;

	struct LALR_TABLE_ITEM {
		bool			isShift;			// ,  

		union {
			int			state;				//   -  ,   TOKEN_EOF,    ( )
			RULE_INDEX	furl;				//  ,   
		};
	};
	typedef std::map<RULE::ITEM, LALR_TABLE_ITEM> LALR_TABLE_LINE;
	typedef std::vector<LALR_TABLE_LINE>	LALR_TABLE;

	typedef std::set<LALR_ITEM>				LALR_ITEM_SET;
	typedef std::vector<LALR_ITEM_SET>		LALR_STATE_ARRAY;

	typedef std::set<RULE_INDEX>			RULE_INDEX_SET;

	typedef std::map<std::string, P_AUTO_FUNCTION> AUTOFUNCTION_MAP;
	typedef std::map<P_AUTO_FUNCTION, std::string> AUTOFUNCTION_REV_MAP;

	typedef std::map<RULE_INDEX, INT_SET>	PREFER_MAP;
	typedef std::map<int, PRIORITY>			TERM_PRIORITY_MAP;
	typedef std::map<RULE_INDEX, PRIORITY>	RULE_PRIORITY_MAP;

private:
	std::string			TransformNameToAlias(const std::string & name) const;

	void				PrintRule			(std::ostream & output, int nontermId, int ruleId, bool printEoln = false) const { PrintRule(output, nonterminal[nontermId].name, nonterminal[nontermId].rules[ruleId].items, printEoln); }
	void				PrintRule			(std::ostream & output, const std::string & name, const RULE::ITEM_ARRAY & ruleItems, bool printEoln = false) const;

	void				AddNonterminal		(const std::string & name);
	void				AddTerminal			(const std::string & name);

	void				Closing				(const LALR_ITEM_SET & state, LALR_ITEM_SET & resultState) const;

	template <class T>
	static bool			IsSubset			(const std::set<T> & left, const std::set<T> & right);	

	void				PrintLalrItem		(std::ostream & output, const LALR_ITEM & ruleItem, bool printEoln = false) const;

	typedef std::map<RULE_INDEX, int> RULE_NUMBER_MAP;
	void				PrintActionTableItem(const LALR_TABLE_LINE & line, RULE::ITEM item, RULE_NUMBER_MAP & ruleNumber, std::ostream & code);

	static void			PrintStringWithPrefix(std::ostream & output, const std::string & prefix, const std::string & text);

	std::string			MakeAnonymousNontermName(void) const;
	std::string			RegisterAutoFunction(const AUTO_FUNCTION & af);

	void				CheckUnreachableNonterminals(const std::string & startNonterm, std::ostream & error, bool & hasError);
	void				MakeTerminalStep	(const LALR_ITEM_SET & state, int termId, int & minindex, int stateIndex);
	void				MakeNonterminalStep	(const LALR_ITEM_SET & state, int nontermId, int & minindex, int stateIndex);
	void				ProcessItemReduce	(const LALR_ITEM & item, int stateId, const LALR_ITEM_SET & state, std::ostream & error, bool & hasError);

	bool				GetPrefer			(int nontermId, int ruleId, INT_SET & specification, bool preferShift) const;
	void				RegisterPrefer		(int nontermId, int ruleId, const INT_SET & specification, bool preferShift);

	enum PREFER {
		PREFER_NONE,
		PREFER_SHIFT,
		PREFER_REDUCE,
		PREFER_OLD_REDUCE,
	};

	PREFER				FindPrefer			(RULE_INDEX index, LALR_TABLE_ITEM oldItem, int termId, std::ostream & error);
	static bool			IsPrefer			(RULE_INDEX index, int termId, PREFER_MAP & prefer);
	static int			FindRule			(const RULE_ARRAY & rules, const RULE & toFind);

	bool				RegisterOpPriority	(int termId, bool left, int priority);
	bool				RegisterRulePriority(RULE_INDEX rule, bool left, int priority);

	PRIORITY			FindPriority		(RULE_INDEX index);
	inline std::string	ConvertToLiteral	(const std::string & value);
	inline std::string	GlobalTypeName		(const std::string & value);

private:
	LOADER				loader;

	std::string			startNonterm;

	// GENERATOR DATA
	STRING_INT_MAP		terminalNumber;		//    
	TERMINAL_ARRAY		terminal;			//  

	STRING_INT_MAP		nonterminalNumber;	//    
	NONTERMINAL_ARRAY	nonterminal;		//  

	LALR_STATE_ARRAY	lalrStates;			//  LALR-
	LALR_TABLE			lalrTable;			// LALR- 

	PREFER_MAP			preferShift;		//  , ""  
	PREFER_MAP			preferReduce;		//  , ""    

	AUTOFUNCTION_MAP	autoFuncs;			// automatic function set for parametrized calls like Print("Hello world")
	AUTOFUNCTION_REV_MAP autoFuncsRev;		// reverse map for same functions look up

	TERM_PRIORITY_MAP	priorityOp;			// operator priority
	RULE_PRIORITY_MAP	priorityRule;		// rule priority

	// NAMING DATA
	std::string			name;				// generated compiler name
	std::string			nameImpl;			// generated compiler implementation name
	std::string			termPrefix;			//   alias- 
	std::string			nontermPrefix;		//   alias- 

	std::string			headerName;			// target header filename
	std::string			codeName;			// target code filename
	std::string			implName;			// implementation name

	// LEX DATA
	bool				enableLex;			// enable lex
	std::string			nameLex;			// generated lex name

	// MESSAGES
	bool				processMsgOnly;

	typedef std::map<std::string, MESSAGE> MESSAGE_MAP;
	bool				enableMessages;
	MESSAGE_MAP			messages;

	STRING_STRING_MAP	aliasTA;			// terminal-alias:      ,  
	STRING_INT_MAP		aliasAT;			// alias-terminal:      ,  
	mutable int			aliasCounter;		// counter for automatic name generation
	mutable STRING_SET	aliasSet;			// alias set for quick unique search
};

/*************************************************************************/
/* GENERATOR::IsSubset()                                                 */
/*                                                                       */
/*************************************************************************/
template <class T>
bool GENERATOR::IsSubset(const std::set<T> & left, const std::set<T> & right)
{
	for (std::set<T>::const_iterator iterLeft = left.begin(); iterLeft != left.end(); ++iterLeft) {
		if (right.find(*iterLeft) == right.end()) {
			return false;
		}
	}
	return true;
}

} // namespace ecg

#endif // ECG_GENERATOR_H
