
/*************************************************************************/
/* ECG_INTSET.HPP                                             07.08.2006 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

#ifndef ECG_INTSET_H
#error ecg_intset.h must be included first
#endif

/*************************************************************************/
/* INTERVAL::queryInter()                                                */
/*                                                                       */
/*************************************************************************/
template <class T>
bool INTERVAL<T>::queryInter(const INTERVAL<T> & a, INTERVAL<T> & res) const
{
	if (start < a.start) {		// [t1.. [a1..
		if (end < a.start) {	// [t1 t2] [a1..
			return false;
		}

		res.start		= a.start;

		if (end < a.end) {		// [t1 [a1 t2] a2]
			res.end		= end;
		} else {				// [t1 [a1 a2] t2]
			res.end		= a.end;
		}

		return true;
	} else {					// [a1.. [t1..
		if (a.end < start) {	// [a1 a2] [t1..
			return false;
		}

		res.start		= start;

		if (a.end < end) {		// [a1 [t1 a2] t2]
			res.end		= a.end;
		} else {				// [a1 [t1 t2] a2]
			res.end		= end;
		}

		return true;
	}
}

/*************************************************************************/
/* INTERVAL::querySub()                                                  */
/*                                                                       */
/*************************************************************************/
template <class T>
int INTERVAL<T>::querySub(const INTERVAL<T> & a, INTERVAL<T> & res, INTERVAL<T> & res2) const
{
	if (start < a.start) {		// [t1.. [a1..
		if (end < a.start) {	// [t1 t2] [a1..
			res = *this;
			return 1;
		}

		// first part
		res.start	= start;			// [t1..  [a1..
		res.end		= a.start - 1;		// [r1 r1]

		// check for one part
		if (end <= a.end) {
			return 1;
		}

		// [t1    [a1 a2]    t2]
		// [r1 r1]       [r2 r2]
		res2.start  = a.end + 1;
		res2.end	= end;
		return 2;
	}

	// [a1 a2] [t1..
	if (a.end < start) {
		res = *this;
		return 1;
	}

	// [a1 [t1 a2]    t2]
	//            [r1 r1]
	if (a.end < end) {		
		res.start	= a.end + 1;
		res.end		= end;
		return 1;
	}
	
	// [a1 [t1 t2] a2]
	return 0;
}

/*************************************************************************/
/* operator + ()                                                         */
/*                                                                       */
/*************************************************************************/
template <class T>
INTSET<T> INTSET<T>::operator + (const INTSET_T & a) const
{
	INTSET_T res = *this;
	return res += a;
}

/*************************************************************************/
/* operator += ()                                                        */
/*                                                                       */
/*************************************************************************/
template <class T>
INTSET<T> & INTSET<T>::operator += (const INTSET_T & a)
{
	if (negative && a.negative) {
		// ~a + ~b = ~(a * b)
		mul(a);
		return *this;
	}

	if (negative) {
		// ~a + b = ~(a * ~b) = ~(a - b)
		sub(a);
		return *this;
	}

	if (a.negative) {
		// a + ~b = ~(~a * b) = ~(b - a)
		INTSET_T tmp = a;
		tmp.sub(*this);
		intervals.swap(tmp.intervals);
		return negate();
	}

	// a + b
	add(a);
	return *this;
}

/*************************************************************************/
/* INTSET<T>::add()                                                      */
/*                                                                       */
/*************************************************************************/
template <class T>
void INTSET<T>::add(const INTSET_T & a)
{
	INTERVAL_ARRAY resIntervals;
	std::copy(intervals.begin(), intervals.end(), std::back_inserter(resIntervals));
	std::copy(a.intervals.begin(), a.intervals.end(), std::back_inserter(resIntervals));

	intervals.swap(resIntervals);
	optimize();
}

/*************************************************************************/
/* operator *= ()                                                        */
/*                                                                       */
/*************************************************************************/
template <class T>
INTSET<T> & INTSET<T>::operator *= (const INTSET_T & a)
{
	if (negative && a.negative) {
		// ~a * ~b = ~(a + b)
		add(a);
		return *this;
	}

	if (negative) {
		// ~a * b = b - a;
		INTSET_T tmp = a;
		tmp.sub(*this);
		intervals.swap(tmp.intervals);
		return negate();
	}

	if (a.negative) {
		// a * ~b = a - b;
		sub(a);
		return *this;
	}

	// a * b
	mul(a);
	return *this;
}

/*************************************************************************/
/* INTSET<T>::mul()                                                      */
/*                                                                       */
/*************************************************************************/
template <class T>
void INTSET<T>::mul(const INTSET_T & a)
{
	INTERVAL_T inter;
	INTERVAL_ARRAY res;
	// assume everything optimized
	INTERVAL_ARRAY::const_iterator cur = a.intervals.begin();
	for (INTERVAL_ARRAY::const_iterator left = intervals.begin(); left != intervals.end(); ++left) {
		for (INTERVAL_ARRAY::const_iterator right = cur; right != a.intervals.end(); ++right) {
			if (!left->queryInter(*right, inter)) {
				continue;
			}
			res.push_back(inter);
			cur = right;
		}
	}

	intervals.swap(res);
	optimize();
}

/*************************************************************************/
/* operator * ()                                                         */
/*                                                                       */
/*************************************************************************/
template <class T>
INTSET<T> INTSET<T>::operator * (const INTSET_T & a) const
{
	INTSET_T res = *this;
	return res *= a;
}

/*************************************************************************/
/* operator -=()                                                         */
/*                                                                       */
/*************************************************************************/
template <class T>
INTSET<T> & INTSET<T>::operator -= (const INTSET_T & a)
{
	if (negative && a.negative) {
		// ~a - ~b = ~a * b = b - a
		INTSET_T tmp = a;
		tmp.sub(*this);
		intervals.swap(tmp.intervals);
		return negate();
	}

	if (negative) {
		// ~a - b = ~a * ~b = ~(a + b)
		add(a);
		return *this;
	}

	if (a.negative) {
		// a - ~b = a * b
		mul(a);
		return *this;
	}

	// a - b
	sub(a);
	return *this;
}

/*************************************************************************/
/* INTSET<T>::sub()                                                      */
/*                                                                       */
/*************************************************************************/
template <class T>
void INTSET<T>::sub(const INTSET_T & a)
{
	INTERVAL_ARRAY res;
	INTSET_T inter, inter2;
	INTERVAL_T r1, r2;
	for (INTERVAL_ARRAY::const_iterator left = intervals.begin(); left != intervals.end(); ++left) {
		inter.clear();
		inter.add(*left);
		for (INTERVAL_ARRAY::const_iterator right = a.intervals.begin(); right != a.intervals.end(); ++right) {
			int qres = left->querySub(*right, r1, r2);
			if (qres > 0) {
				inter2.clear();
				inter2.add(r1);
				if (qres > 1) {
					assert(qres == 2);
					inter2.add(r2);
				}
				inter *= inter2;
			} else {
				assert(qres == 0);
				inter.clear();
				break;
			}
		}
		std::copy(inter.intervals.begin(), inter.intervals.end(), std::back_inserter(res));
	}

	intervals.swap(res);
	optimize();
}

/*************************************************************************/
/* operator -()                                                          */
/*                                                                       */
/*************************************************************************/
template <class T>
INTSET<T> INTSET<T>::operator - (const INTSET_T & a) const
{
	INTSET_T res = *this;
	return res -= a;
}

/*************************************************************************/
/* INTSET<T>::negate()                                                   */
/*                                                                       */
/*************************************************************************/
template <class T>
INTSET<T> & INTSET<T>::negate(void)
{
	negative = !negative;
	return *this;
}

/*************************************************************************/
/* operator ~()                                                          */
/*                                                                       */
/*************************************************************************/
template <class T>
INTSET<T> INTSET<T>::operator ~ (void) const
{
	INTSET_T res = *this;
	return res.negate();
}

/*************************************************************************/
/* INTSET<T>::optimize()                                                 */
/*                                                                       */
/*************************************************************************/
template <class T>
INTSET<T> & INTSET<T>::optimize(void)
{
	if (intervals.empty()) {
		return *this;
	}

	INTERVAL_ARRAY res;

	std::sort(intervals.begin(), intervals.end());

	INTERVAL_T inter = intervals[0];
	INTERVAL_T un;
	for (INTERVAL_ARRAY::const_iterator iter = intervals.begin() + 1; iter != intervals.end(); ++iter) {
		if (inter.end + 1 >= iter->start) {
			inter.end = max(inter.end, iter->end);
			continue;
		}

		res.push_back(inter);
		inter = *iter;
	}
	res.push_back(inter);

	intervals.swap(res);

	return *this;
}

/*************************************************************************/
/* operator ==()                                                         */
/*                                                                       */
/*************************************************************************/
template <class T>
bool INTSET<T>::operator == (const INTSET_T & a) const
{
	if (negative != a.negative || intervals.size() != a.intervals.size()) {
		return false;
	}

	for (int i = 0; (size_t)i < intervals.size(); ++i) {
		if (intervals[i] != a.intervals[i]) {
			return false;
		}
	}

	return true;
}

/*************************************************************************/
/* operator <()                                                          */
/*                                                                       */
/*************************************************************************/
template <class T>
bool INTSET<T>::operator < (const INTSET_T & a) const
{
	for (
		INTERVAL_ARRAY::const_iterator i1 = intervals.begin(), i2 = a.intervals.begin();
		i1 != intervals.end() && i2 != a.intervals.end();
		++i1, ++i2)
	{
		if (*i1 < *i2) {
			return true;
		} else if (*i1 != *i2) {
			return false;
		}
	}

	if (i1 == intervals.end() && i2 != a.intervals.end()) {
		return true;
	}

	return false;
}

/*************************************************************************/
/* operator <<()                                                         */
/*                                                                       */
/*************************************************************************/
template <class T>
std::ostream & operator << (std::ostream & output, const INTERVAL<T> & a)
{
	return output << "[" << a.start << "," << a.end << "]";
}

/*************************************************************************/
/* operator <<()                                                         */
/*                                                                       */
/*************************************************************************/
template <class T>
std::ostream & operator << (std::ostream & output, const INTSET<T> & a)
{
	if (a.negative) {
		output << "not ";
	}
	if (a.intervals.empty()) {
		output << "<empty>";
	} else {
		std::copy(a.intervals.begin(), a.intervals.end(), ostream_iterator<INTERVAL<T> > (output, " "));
	}
	return output;
}

//
// End-of-file ECG_INTSET.HPP
//
