
/*************************************************************************/
/* ECG_LEX_IMPL.CPP                                           08.08.2006 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

#include "stdafx.h"
#include "ecg_generator.h"

using namespace std;
using namespace ecg;

/*************************************************************************/
/* LEX::GetChar()                                                        */
/*                                                                       */
/*************************************************************************/
bool LEX::GetChar(CHAR_T & ch)
{
	if (regexpPos == regexp.size()) {
		return false;
	}

	ch = regexp[regexpPos++];
	return true;
}

/*************************************************************************/
/* LEX::LEX()                                                            */
/*                                                                       */
/*************************************************************************/
LEX::LEX(void)
	: enablePresets(false)
	, errorOnRedefinition(true)
	, commentLine(TOKEN_EOF)
	, commentBlockBegin(TOKEN_EOF)
	, commentBlockEnd(TOKEN_EOF)
	, defaultWasModified(false)
{}

/*************************************************************************/
/* LEX::Init()                                                           */
/*                                                                       */
/*************************************************************************/
void LEX::Init(int tkn, const std::string & rexp, PRIORITY priority, int line, std::ostream & error, bool errOnRedefinition /*= true*/)
{
	Init(line, error);
	regexp			= rexp;
	regexpPos		= 0;
	token			= tkn;
	errorOnRedefinition = errOnRedefinition;
	userPriority	= priority;
}

/*************************************************************************/
/* LEX::Init()                                                           */
/*                                                                       */
/*************************************************************************/
void LEX::Init(int line, std::ostream & error)
{
	this->line		= line;
	this->error		= &error;
}

/*************************************************************************/
/* LEX::GetError()                                                       */
/*                                                                       */
/*************************************************************************/
ostream & LEX::GetError(void) const
{
	assert(error);
	return *error;
}

/*************************************************************************/
/* LEX::Line()                                                           */
/*                                                                       */
/*************************************************************************/
string LEX::Line(void) const
{
	ostringstream res;
	res << line << "(" << regexpPos << ")";
	return res.str();
}

/*************************************************************************/
/* LEX::NewNode()                                                        */
/*                                                                       */
/*************************************************************************/
int LEX::NewNode(void)
{
	int res = (int)nodes.size();
	nodes.push_back(NODE());
	return res;
}

/*************************************************************************/
/* LEX::SimpleRegexp()                                              
/*                                                                       
/* e			-> $char;								<<< SimpleRegexp >>>
/*************************************************************************/
void LEX::SimpleRegexp(const std::string & ch, P<REGEXP> & res)
{
	assert(ch.size() == 1);

	// allocate result regexp
	res = new REGEXP();

	// fill start and final node numbers
	res->start	= NewNode();
	res->end	= NewNode();

	// edge [start -(ch)-> end]
	nodes[res->start].AddEdge(CHAR_INTSET().add(ch[0]), res->end);
}

/*************************************************************************/
/* LEX::FixedCharSet()                                                   */
/*                                                                       */
/*************************************************************************/
bool LEX::FixedCharSet(char ch, CHAR_INTSET & charSet) const
{
	switch (ch) {
		case 'c': charSet.clear(); charSet.negate(); break;
		case 's': charSet.add(0, ' ').optimize();	break;
		case 'S': charSet.add(0, ' ').negate().optimize(); break;
		case 'd': charSet.add('0', '9').optimize();	break;
		case 'D': charSet.add('0', '9').add(0, ' ').optimize().negate(); break;
		case 'b': charSet.add(' ').optimize();		break;
		case 't': charSet.add('\t').optimize();		break;
		case 'n': charSet.add('\n').optimize();		break;
		case 'a': charSet.add('a', 'z').add('A', 'Z').optimize(); break;
		case 'y': charSet.add('\\').add('\"').add('\'')
						 .add(':').add(';').add(',').add('.')
						 .add('?').add('!').add('~').add('#')
						 .add('%').add('^').add('&').add('|')
						 .add('+').add('-').add('*').add('/')
						 .add('(').add(')').add('[').add(']').add('{').add('}')
						 .add('<').add('>').add('=').optimize(); break;

		default:  charSet.add(ch).optimize();		return !isalpha(char_traits<char>::to_int_type(ch));
	}

	return true;
}

/*************************************************************************/
/* LEX::FixedSetRegexp()                                            
/*                                                                       
/* e			-> SLASH $char;							<<< FixedSetRegexp >>>
/*************************************************************************/
void LEX::FixedSetRegexp(const std::string & ch, P<REGEXP> & res)
{
	assert(ch.size() == 1);

	// alloc result regexp
	res = new REGEXP();
	res->start	= NewNode();
	res->end	= NewNode();

	CHAR_INTSET cond;
	if (!FixedCharSet(ch[0], cond)) {
		messageTable["name"]	= ch;
		FatalError(MESSAGE_LEX1012);
	}

	nodes[res->start].AddEdge(cond, res->end);
}

/*************************************************************************/
/* LEX::SingleChar()                                                
/*                                                                       
/* s			-> $char;								<<< SingleChar >>>
/*************************************************************************/
void LEX::SingleChar(const string & ch, P<CHARS> & res)
{
	res = new CHARS();
	res->add(ch[0]);
}

/*************************************************************************/
/* LEX::RangeChars()                                                
/*                                                                       
/* s			-> $char "-" $char;						<<< RangeChars >>>
/*************************************************************************/
void LEX::RangeChars(const std::string & ch, const std::string & ch2, P<CHARS> & res)
{
	res = new CHARS();
	res->add(ch[0], ch2[0]);
}

/*************************************************************************/
/* LEX::FixedCharSet()                                              
/*                                                                       
/* s			-> "\\" $char;							<<< FixedCharSet >>>
/*************************************************************************/
void LEX::FixedCharSet(const std::string & ch, P<CHARS> & res)
{
	assert(ch.size() == 1);
	res = new CHARS();

	if (!FixedCharSet(ch[0], *res)) {
		messageTable["name"]	= ch;
		FatalError(MESSAGE_LEX1012);
	}
}

/*************************************************************************/
/* LEX::CharsUnion()                                                
/*                                                                       
/* set			-> set s;								<<< CharsUnion >>>
/*************************************************************************/
void LEX::CharsUnion(P<CHARS> s, P<CHARS> s2, P<CHARS> & res)
{
	res = s;
	*res += *s2;
}

/*************************************************************************/
/* LEX::SetRegexp()                                                 
/*                                                                       
/* e			-> "{" set "}";							<<< SetRegexp >>>
/*************************************************************************/
void LEX::SetRegexp(P<CHARS> chars, P<REGEXP> & res)
{
	res = new REGEXP();
	res->start	= NewNode();
	res->end	= NewNode();

	nodes[res->start].AddEdge(*chars, res->end);
}

/*************************************************************************/
/* LEX::NegativeSetRegexp()                                         
/*                                                                       
/* e			-> "{" set "^" set "}";					<<< NegativeSetRegexp >>>
/*************************************************************************/
void LEX::NegativeSet(P<CHARS> a, P<CHARS> b, P<CHARS> & res)
{
	// {^a} same as {\c^a}
	if (a->empty()) {
		a->negate();
	}

	res = a;
	*res -= *b;
}

/*************************************************************************/
/* LEX::ConcatRegexp()                                              
/*                                                                       
/* e			-> e e;				prior(concat)		<<< ConcatRegexp >>>
/*************************************************************************/
void LEX::ConcatRegexp(P<REGEXP> a, P<REGEXP> b, P<REGEXP> & res)
{
	// a: as -> ae
	// b: bs -> be
	// res: as -> ae -> bs -> be
	nodes[a->end].AddEdge(b->start);

	// use first argument for result of concatenation
	res = a;
	res->end = b->end;
}

/*************************************************************************/
/* LEX::AnyRegexp()                                                 
/*                                                                       
/* e			-> e "*";								<<< AnyRegexp >>>
/*************************************************************************/
void LEX::AnyRegexp(P<REGEXP> a, P<REGEXP> & res)
{
	res = new REGEXP();
	res->start	= NewNode();

	// '*' means [rs -> as], [ae = re]
	nodes[res->start].AddEdge(a->start);
	res->end	= a->end;

	// and [rs -> re], [re -> rs]
	nodes[res->start].AddEdge(res->end);
	nodes[res->end].AddEdge(res->start);
}

/*************************************************************************/
/* LEX::ManyRegexp()                                                
/*                                                                       
/* e			-> e "+";								<<< ManyRegexp >>>
/*************************************************************************/
void LEX::ManyRegexp(P<REGEXP> a, P<REGEXP> & res)
{
	res = a;

	// '+' means [ae -> as]
	nodes[res->end].AddEdge(res->start);
}

/*************************************************************************/
/* LEX::OrRegexp()                                                  
/*                                                                       
/* e			-> e "|" e;								<<< OrRegexp >>>
/*************************************************************************/
void LEX::OrRegexp(P<REGEXP> a, P<REGEXP> b, P<REGEXP> & res)
{
	res = new REGEXP();
	res->start = NewNode();

	// '|' means [rs -> as], [rs -> bs]
	nodes[res->start].AddEdge(a->start);
	nodes[res->start].AddEdge(b->start);

	// and [re = ae], [be -> re]
	res->end = a->end;
	nodes[b->end].AddEdge(res->end);
}

/*************************************************************************/
/* LEX::YesNotRegexp()                                                   */
/*                                                                       */
/*************************************************************************/
void LEX::YesNotRegexp(P<REGEXP> a, P<REGEXP> & res)
{
	res = new REGEXP();
	res->start	= NewNode();
	// '[]' means [rs -> as], [ae = re]
	nodes[res->start].AddEdge(a->start);
	res->end	= a->end;

	// and [rs -> re]
	nodes[res->start].AddEdge(res->end);
}

/*************************************************************************/
/* LEX::SetPreset()                                                      */
/*                                                                       */
/*************************************************************************/
void LEX::SetPreset(void)
{
	SetPreset(PRESET_SET);
}

/*************************************************************************/
/* LEX::DefPreset()                                                      */
/*                                                                       */
/*************************************************************************/
void LEX::DefPreset(void)
{
	SetPreset(PRESET_DEFAULT);
}

/*************************************************************************/
/* LEX::SlashPreset()                                                    */
/*                                                                       */
/*************************************************************************/
void LEX::SlashPreset(void)
{
	PushPreset(PRESET_SLASH);
}

/*************************************************************************/
/* LEX::RestorePreset()                                                  */
/*                                                                       */
/*************************************************************************/
void LEX::RestorePreset(string & res)
{
	res		= TokenValue();
	PopPreset();
}

/*************************************************************************/
/* LEX::PrintMessage()                                                   */
/*                                                                       */
/*************************************************************************/
void LEX::PrintMessage(int messageIndex)
{
	string msg = LEX_GetMessageByIndex(messageIndex)->format;
	msg = "Line {line}: " + msg;
	messageTable["line"] = Line();
	GetError() << Compose(msg, messageTable) << endl;
}

/*************************************************************************/
/* LEX::FatalError()                                                     */
/*                                                                       */
/*************************************************************************/
void LEX::FatalError(int messageIndex)
{
	PrintMessage(messageIndex);
	compilerState.set(CST_FATAL_ERROR);
}

/*************************************************************************/
/* LEX::SetFinalRegexp()                                                 */
/*                                                                       */
/*************************************************************************/
void LEX::SetFinalRegexp(P<REGEXP> a)
{
	// check gen's terminal redifinition
	if (genTerms[genTermsNmb[token]].hasRegexp && errorOnRedefinition) {
		messageTable["name"] = genTerms[token].name;
		FatalError(MESSAGE_LEX1000);
		return ;
	}

    genTerms[genTermsNmb[token]].hasRegexp = true;
	nodes[a->end].token = token;
	startNodes.insert(a->start);

	// calc priority
	PRIORITY prior = userPriority;
	if (prior == PRIORITY_NONE) {
		prior = PRIORITY_HIGH;
		int node = a->start;
		bool isEnd = false;
		do {
			// check if it's last node
			isEnd = (node == a->end);

			int edgeCount = (int)nodes[node].edges.size();

			// check: simple node must has one edge, final node - no edges
			if (!isEnd && edgeCount != 1
			||   isEnd && edgeCount != 0) {
				prior = PRIORITY_LOW;
				break;
			}

			if (!isEnd) {
				assert(edgeCount == 1);
				CHARS & cond = nodes[node].edges[0].cond;

				// if rule has edge with non-simple condition (allow multiple matching for this regexp)
				if (cond.negative
				||  cond.intervals.size() > 1
				||	cond.intervals.size() == 1 && cond.intervals[0].start != cond.intervals[0].end) {
					prior = PRIORITY_NORMAL;
				}

				node = nodes[node].edges[0].next;
			}
		} while (!isEnd);
	}

	nodes[a->end].prior = prior;
}

/*************************************************************************/
/* LEX::AddOrthogonal()                                                  */
/*                                                                       */
/*************************************************************************/
void LEX::AddOrthogonal(const CHARS & a)
{
	CHARS tunion;
	CHARS_ARRAY res;
	res.reserve(terminals.size() + 2);

	CHARS sub;
	CHARS mul;
	for (CHARS_ARRAY::const_iterator Tk = terminals.begin(); Tk != terminals.end(); ++Tk) {
		// calc Tk * a
		mul = *Tk;
		mul *= a;

		// Tk * a == Tk
		// Tk * a == 0
		if (mul.empty() || mul == *Tk) {
			res.push_back(*Tk);
			tunion += *Tk;
			continue;
		}

		// add new terminals
		// Tk * a
		res.push_back(mul);
		tunion += mul;

		// Tk - a
		sub = *Tk;
		sub -= a;
		assert(!sub.empty());
		res.push_back(sub);
		tunion += sub;
	}

	// calc a - sum(Tk)
	sub = a;
	sub -= tunion;

	// if (a-sum(Tk)) == 0 -> a can be composed by Tks
	if (!sub.empty()) {
		res.push_back(sub);
	}

	// result
	terminals.swap(res);
}

/*************************************************************************/
/* LEX::Orthogonalize()                                                  */
/*                                                                       */
/*************************************************************************/
void LEX::Orthogonalize(void)
{
	terminals.clear();

	// calc orthogonalized terminals
	for (NODE_ARRAY::const_iterator node = nodes.begin(); node != nodes.end(); ++node) {
		for (EDGE_ARRAY::const_iterator edge = node->edges.begin(); edge != node->edges.end(); ++edge) {
			AddOrthogonal(edge->cond);
		}
	}

	// skip symbols
	CHARS ch;
	FixedCharSet('s', ch);
	AddOrthogonal(ch);

	// break symbols
	FixedCharSet('y', ch);
	AddOrthogonal(ch);

	// calc condition from terminals
	for (NODE_ARRAY::iterator node = nodes.begin(); node != nodes.end(); ++node) {
		for (EDGE_ARRAY::iterator edge = node->edges.begin(); edge != node->edges.end(); ++edge) {
			// fill terminal condition
			edge->condTerm.clear();
			for (int k=0; (size_t)k < terminals.size(); ++k) {
				if (!(edge->cond * terminals[k]).empty()) {
					edge->condTerm.insert(k);
				}
			}
		}
	}
}

/*************************************************************************/
/* LEX::Closing()                                                        */
/*                                                                       */
/*************************************************************************/
void LEX::Closing(int node, INT_INT_MAP & edges, INT_SET & states)
{
	if (states.find(node) != states.end()) {
		return ;
	}

	// mark this state already processed
	states.insert(node);

	// check other nodes
	for (EDGE_ARRAY::const_iterator edge = nodes[node].edges.begin(); edge != nodes[node].edges.end(); ++edge) {
		assert(edge->condTerm.empty() == edge->cond.empty());

		// store terminals
		for (INT_SET::const_iterator term = edge->condTerm.begin(); term != edge->condTerm.end(); ++term) {
			// set empty edge
			edges[*term] = -1;
		}

		// epsilon-edge
		if (edge->condTerm.empty() && states.find(edge->next) == states.end()) {
			// recursive closing
			Closing(edge->next, edges, states);
		}
	}
}

/*************************************************************************/
/* LEX::Closing()                                                        */
/*                                                                       */
/*************************************************************************/
void LEX::Closing(STATE & state)
{
	state.edges.clear();
	INT_SET nodes;
	for (INT_SET::const_iterator nd = state.nodes.begin(); nd != state.nodes.end(); ++nd) {
		Closing(*nd, state.edges, nodes);
	}

	state.nodes.swap(nodes);

	// optimization
	nodes.clear();
	for (INT_SET::const_iterator nd = state.nodes.begin(); nd != state.nodes.end(); ++nd) {
		// final node
		if (this->nodes[*nd].token != TOKEN_EOF) {
			nodes.insert(*nd);
			continue;
		}

		// has non-eps edges
		bool hasNonEpsEdge = false;
		for (EDGE_ARRAY::const_iterator edge = this->nodes[*nd].edges.begin(); edge != this->nodes[*nd].edges.end(); ++edge) {
			if (!edge->cond.empty()) {
				hasNonEpsEdge = true;
				break;
			}
		}
		if (hasNonEpsEdge) {
			nodes.insert(*nd);
			continue;
		}

		// otherwise skip it
	}
	state.nodes.swap(nodes);
}

/*************************************************************************/
/* LEX::RetreiveToken()                                                  */
/*                                                                       */
/*************************************************************************/
bool LEX::RetreiveToken(const STATE & state, int & token, const BOOL_ARRAY & genTerms, bool skipConflicts)
{
	// check for final state
	token			= TOKEN_EOF;
	PRIORITY prior	= PRIORITY_NONE;
	for (INT_SET::const_iterator node = state.nodes.begin(); node != state.nodes.end(); ++node) {
		// skip non-final node
		if (nodes[*node].token == TOKEN_EOF) {
			continue;
		}

		// skip disabled terminals
		assert(genTermsNmb.find(nodes[*node].token) != genTermsNmb.end());
		if (!genTerms[genTermsNmb[nodes[*node].token]]) {
			continue;
		}

		// check for first token
		if (token == TOKEN_EOF) {
			token = nodes[*node].token;
			prior = nodes[*node].prior;
			continue;
		}

		// check for token's priority
		if (token != nodes[*node].token) {
			if (prior > nodes[*node].prior) {
				// just skip this
				continue;
			}

			if (prior < nodes[*node].prior) {
				// override because priority
				token = nodes[*node].token;
				prior = nodes[*node].prior;
				continue;
			}

			// skip conflicts when calculating table and using presets
			if (!skipConflicts) {
				// handle error
				messageTable["name"]	= this->genTerms[genTermsNmb[token]].name;
				messageTable["name2"]	= this->genTerms[genTermsNmb[nodes[*node].token]].name;
				PrintMessage(MESSAGE_LEX1001);
				return false;
			}
		}
	}

	return true;
}

/*************************************************************************/
/* LEX::CalcTable()                                                      */
/*                                                                       */
/*************************************************************************/
bool LEX::CalcTable(void)
{
	// do orthogonalization
	Orthogonalize();

	// create first state
	table.clear();
	P_STATE firstState = new STATE();
	table.push_back(firstState);
	table[0]->nodes = startNodes;
	searchTable[firstState] = 0;

	Closing(*table[0]);

	// by default all tokens enabled
	BOOL_ARRAY genTerms(this->genTerms.size(), true);

	P_STATE newState = new STATE();
	int curState = 0;
	do {
		STATE & state = *table[curState];

		if (!RetreiveToken(state, state.token, genTerms, enablePresets)) {
			return false;
		}

		GENTERM t = this->genTerms[genTermsNmb[state.token]];

		int a = 0;

		for (INT_INT_MAP::const_iterator term = state.edges.begin(); term != state.edges.end(); ++term) {
			// make step for this terminal
			newState->nodes.clear();
			newState->edges.clear();
			MakeStep(state.nodes, term->first, newState->nodes);
			assert(!newState->nodes.empty());

			// make closing
			Closing(*newState);

			// try to find existing state
			STATE_INT_MAP::const_iterator existingState = searchTable.find(newState);
			if (existingState != searchTable.end()) {
				state.edges[term->first] = existingState->second;
				continue;
			}

			// add new state
			int stateNmb = (int)table.size();
			table.push_back(newState);
			searchTable[newState] = stateNmb;

			// add table edge
			state.edges[term->first] = stateNmb;

			// prepare for future use
			newState.reset(new STATE());
		}

		++curState;
	} while ((size_t)curState < table.size());

	// commit presets
	if (!CommitPresets()) {
		PrintMessage(MESSAGE_LEX1002);
		return false;
	}

	return true;
}

/*************************************************************************/
/* operator <()                                                          */
/*                                                                       */
/*************************************************************************/
bool LEX::STATE::operator < (const STATE & a) const
{
	INT_SET::const_iterator it1 = nodes.begin(), it2 = a.nodes.begin();
	for (;
		it1 != nodes.end() && it2 != a.nodes.end();
		++it1, ++it2)
	{
		if (*it1 < *it2) {
			return true;
		} else if (*it1 != *it2) {
			return false;
		}
	}

	if (it1 == nodes.end() && it2 != a.nodes.end()) {
		return true;
	}

	return false;
}

/*************************************************************************/
/* LEX::MakeStep()                                                       */
/*                                                                       */
/*************************************************************************/
void LEX::MakeStep(const INT_SET & nodes, int term, INT_SET & newNodes)
{
	// for each node
	for (INT_SET::const_iterator node = nodes.begin(); node != nodes.end(); ++node) {
		// for each node's edge
		for (EDGE_ARRAY::const_iterator edge = this->nodes[*node].edges.begin(); edge != this->nodes[*node].edges.end(); ++edge) {
			// if edge contains terminal
			if (edge->condTerm.find(term) != edge->condTerm.end()) {
				// put next node to new state's node list
				newNodes.insert(edge->next);
			}
		}
	}
}

/*************************************************************************/
/* LEX::RegisterTerminal()                                               */
/*                                                                       */
/*************************************************************************/
void LEX::RegisterTerminal(int termId, const std::string & name, const std::string & alias)
{
	INT_INT_MAP::const_iterator iter = genTermsNmb.find(termId);
	if (iter != genTermsNmb.end()) {
		assert(genTerms[iter->second].id == termId);
		assert(genTerms[iter->second].name == name);
		assert(genTerms[iter->second].alias == alias);
		return ;
	}

	GENTERM term;
	term.id		= termId;
	term.name	= name;
	term.alias	= alias;

	int nmb = (int)genTerms.size();
	genTerms.push_back(term);
	genTermsNmb[termId] = nmb;
	assert(genTerms.size() == genTermsNmb.size());
}

/*************************************************************************/
/* LEX::StringToChars()                                                  */
/*                                                                       */
/*************************************************************************/
bool LEX::StringToChars(const char * value, int length,  CHARS & ch, bool isSet /*= false*/, bool isSub /*= false*/)
{
	if (length == 0) {
		return true;
	}

	for (int i=0; i < length; ++i) {
		// escape-symbol
		if (value[i] == '\\') {
			if (i + 1 >= length) {
				PrintMessage(MESSAGE_LEX1003);
				return false;
			}
			FixedCharSet(value[i+1], ch);
			++i;
			continue;
		}
		
		// recursive set
		if (value[i] == '{') {
			// check for closing brackets
			int cntr = 1, index = i + 1;
			for (; index < length; ++index) {
				if (value[index] == '{') {
					++cntr;
				} else if (value[index] == '}') {
					--cntr;
				}
				if (cntr == 0) {
					break;
				}
			}
			if (cntr != 0) {
				PrintMessage(MESSAGE_LEX1004);
				return false;
			}

			// subset
			CHARS ch2;
			if (!StringToChars(value + i + 1, index - i - 1, ch2, true)) {
				return false;
			}

			ch.optimize() += ch2;
			i = index;
			continue;
		}

		// check closing brackets
		if (value[i] == '}') {
			PrintMessage(MESSAGE_LEX1005);
			return false;
		}

		// negate
		if (value[i] == '^') {
			if (isSub) {
				messageTable["name"] = value[i];
				PrintMessage(MESSAGE_LEX1006);
				return false;
			}

			CHARS neg;
			if (!StringToChars(value + i + 1, length - i - 1, neg, true, true)) {
				return false;
			}
	
			// {^a} == {\c^a}
			if (ch.empty()) {
				ch.negate();
			}

			ch.optimize() -= neg;
			return true;
		}

		ch.add(value[i]);
	}

	ch.optimize();
	return true;
}

/*************************************************************************/
/* LEX::AddPreset()                                                      */
/*                                                                       */
/*************************************************************************/
bool LEX::AddPreset(const std::string & name, int & presetId)
{
	// register default preset
	// name "default" reserved
	if (presets.empty()) {
		presets.push_back(PRESET());
		presets[0].name = "default";
		presetsNmb["default"] = 0;
		defaultWasModified = false;
	}

	STRING_INT_MAP::const_iterator pres = presetsNmb.find(name);
	if (pres != presetsNmb.end()) {
		presetId = pres->second;
		if (name == "default") {
			if (!defaultWasModified) {
				defaultWasModified = true;
				return true;
			}
		}
		return false;
	}

	presetId = (int)presets.size();
	presets.push_back(PRESET());
	presets[presetId].name = name;
	presetsNmb[name] = presetId;
	assert(presets.size() == presetsNmb.size());

	enablePresets = presets.size() > 1;

	return true;
}

/*************************************************************************/
/* LEX::CalcGenTerms()                                                   */
/*                                                                       */
/*************************************************************************/
void LEX::CalcGenTerms(PRESET & preset, BOOL_ARRAY & gt, int flags)
{
	assert(flags);
	assert(gt.size() == genTerms.size());

	for (int index = 0; (size_t)index < genTerms.size(); ++index) {
		bool found = !(flags & CGT_ONLY_LOCAL);

		// check local preset
		if (genTerms[index].preset != -1) {
			if (flags & CGT_SKIP_LOCAL) {
				found = false;
			} else {
				// just compare presets names because no recursive presets for now
				assert((size_t)genTerms[index].preset < presets.size());
				found = (presets[genTerms[index].preset].name == preset.name);
			}
		}

		if (flags & CGT_FILL) {
			preset.genTerms[index] = found;
		} else if (flags & CGT_ADD) {
			if (found) {
				preset.genTerms[index] = true;
			}
		} else if (flags & CGT_SUB) {
			if (found) {
				preset.genTerms[index] = false;
			}
		} else {
			assert(false);
		}
	}
}

/*************************************************************************/
/* LEX::ProcessPreset()                                                  */
/*                                                                       */
/*************************************************************************/
bool LEX::ProcessPreset(PRESET & preset)
{
	assert(preset.status == PRESET::NOT_PROCESSED);
	preset.status = PRESET::PROCESSING;

	// by default all terms turned off
	assert(preset.genTerms.size() == genTerms.size());
	CalcGenTerms(preset, preset.genTerms, CGT_FILL|CGT_ONLY_LOCAL);

	// by default comment terminals enabled too
	if (commentLine != TOKEN_EOF) {
		assert(genTermsNmb.find(commentLine) != genTermsNmb.end());
		preset.genTerms[genTermsNmb[commentLine]] = true;
	}
	if (commentBlockBegin != TOKEN_EOF && commentBlockEnd != TOKEN_EOF) {
		assert(genTermsNmb.find(commentBlockBegin) != genTermsNmb.end());
		assert(genTermsNmb.find(commentBlockEnd)   != genTermsNmb.end());
		preset.genTerms[genTermsNmb[commentBlockBegin]] = true;
	}

	// process enable/disable commands
	for (COMMAND_ARRAY::const_iterator command = preset.commands.begin(); command != preset.commands.end(); ++command) {
		// enable/disable terminal
		if (command->terminal != TOKEN_EOF) {
			assert(genTermsNmb.find(command->terminal) != genTermsNmb.end());
			preset.genTerms[genTermsNmb[command->terminal]] = command->enable;
			continue;
		}

		// 'all' reserved for all terminals
		if (command->value == ALL_PRESET) {
			if (command->enable) {
				CalcGenTerms(preset, preset.genTerms, CGT_ADD|CGT_SKIP_LOCAL);
			} else {
				CalcGenTerms(preset, preset.genTerms, CGT_SUB|CGT_SKIP_LOCAL);
			}
			continue;
		}

		// 'comment' reserved for comment terminals
		if (command->value == COMMENT_PRESET) {
			if (commentLine != TOKEN_EOF) {
				preset.genTerms[genTermsNmb[commentLine]] = command->enable;
			}
			if (commentBlockBegin != TOKEN_EOF && commentBlockEnd != TOKEN_EOF) {
				preset.genTerms[genTermsNmb[commentBlockBegin]] = command->enable;
			}
			continue;
		}

		// check that preset exists
		if (presetsNmb.find(command->value) == presetsNmb.end()) {
			messageTable["name"]	= command->value;
			messageTable["name2"]	= preset.name;
			PrintMessage(MESSAGE_LEX1007);
			return false;
		}

		// check recursion
		int presetId = presetsNmb[command->value];
		PRESET & prs = presets[presetId];
		if (presets[presetId].status == PRESET::PROCESSING) {
			messageTable["name"]	= prs.name;
			messageTable["name2"]	= preset.name;
			PrintMessage(MESSAGE_LEX1008);
			return false;
		}

		// check unprocessed preset
		if (prs.status == PRESET::NOT_PROCESSED) {
			if (!ProcessPreset(prs)) {
				return false;
			}
		}

		// process single terminal enable
		assert(prs.status == PRESET::PROCESSED);
		for (int index = 0; (size_t)index < prs.genTerms.size(); ++index) {
			if (prs.genTerms[index]) {
				preset.genTerms[index] = command->enable;
			}
		}
	}

	preset.status = PRESET::PROCESSED;

	return true;
}

/*************************************************************************/
/* LEX::MarkProductive()                                                 */
/*                                                                       */
/*************************************************************************/
void LEX::MarkProductive(PRESET & preset, int stateId, REVERSE_GRAPH & reverseGraph)
{
	preset.productive[stateId] = PRESET::PROD_YES;

	int prevCount = (int)reverseGraph[stateId].size();
	for (int prevIndex=0; prevIndex<prevCount; ++prevIndex) {
		int prev = reverseGraph[stateId][prevIndex];
		if (preset.productive[prev] != PRESET::PROD_YES) {
			MarkProductive(preset, prev, reverseGraph);
		}
	}
}

/*************************************************************************/
/* LEX::CommitPresets()                                                  */
/*                                                                       */
/*************************************************************************/
bool LEX::CommitPresets(void)
{
	assert(enablePresets || presets.empty());

	// check for default preset and check that presets are enabled
	enablePresets = true;
	if (presets.empty()) {
		enablePresets = false;
		presets.push_back(PRESET());

		presets[0].isSkip = !defSkip.empty();
		presets[0].skip   = defSkip;
	} else if (!defaultWasModified) {
		// by default 'default' preset should have '+all' command
		int id;
		AddPreset("default", id);
		presets[id].AddCommand(true, ALL_PRESET);
	}

	// prepare presets
	for (PRESET_ARRAY::iterator preset = presets.begin(); preset != presets.end(); ++preset) {
		preset->genTerms.resize(genTerms.size());
		preset->final.resize(table.size(), TOKEN_EOF);
		preset->status = PRESET::NOT_PROCESSED;

		// check default skip
		if (!preset->isSkip && !defSkip.empty()) {
			preset->isSkip	= true;
			preset->skip	= defSkip;
		}
	}

	// fill genTerms bool array
	if (!enablePresets) {
		fill(presets[0].genTerms.begin(), presets[0].genTerms.end(), true);
	} else {
		for (PRESET_ARRAY::iterator preset = presets.begin(); preset != presets.end(); ++preset) {
			if (preset->status == PRESET::NOT_PROCESSED) {
				if (!ProcessPreset(*preset)) {
					return false;
				}
				assert(preset->status == PRESET::PROCESSED);
			}
		}
	}

	// calc final state-terminal table
	for (PRESET_ARRAY::iterator preset = presets.begin(); preset != presets.end(); ++preset) {
		// for each state calc its token
		for (int stateId = 0; (size_t)stateId < table.size(); ++stateId) {
			if (!RetreiveToken(*table[stateId], preset->final[stateId], preset->genTerms, false)) {
				// conflict found
				messageTable["name"]	= preset->name;
				PrintMessage(MESSAGE_LEX1009);
				return false;
			}
		}
	}

	// calc reverse graph for productivity calculation
	REVERSE_GRAPH reverseGraph;
	for (int stateId=0; (size_t)stateId < table.size(); ++stateId) {
		STATE & state = *table[stateId];
		for (INT_INT_MAP::const_iterator edge = state.edges.begin(); edge != state.edges.end(); ++edge) {
			int next = edge->second;
			reverseGraph[next].push_back(stateId);
		}
	}

	// calc productivity for each states
	for (PRESET_ARRAY::iterator preset = presets.begin(); preset != presets.end(); ++preset) {
		preset->productive.resize(table.size());
		fill(preset->productive.begin(), preset->productive.end(), PRESET::PROD_NO);
		for (int stateId = 0; (size_t)stateId < table.size(); ++stateId) {
			if (preset->final[stateId] != TOKEN_EOF) {
				MarkProductive(*preset, stateId, reverseGraph);
			}
		}
	}

	// check comments terminals
	if (commentBlockBegin != TOKEN_EOF || commentBlockEnd != TOKEN_EOF) {
		if (commentBlockBegin == TOKEN_EOF) {
			PrintMessage(MESSAGE_LEX1010);
			return false;
		}
		if (commentBlockEnd == TOKEN_EOF) {
			PrintMessage(MESSAGE_LEX1011);
			return false;
		}
	}

	return true;
}

/*************************************************************************/
/* LEX::RegisterComment()                                                */
/*                                                                       */
/*************************************************************************/
void LEX::RegisterComment(const string & preset, int termId, const string & name, const string & alias)
{
	int presetId;
	if (AddPreset(preset, presetId)) {
		// if preset just added - add '-all' command
		presets[presetId].AddCommand(false, ALL_PRESET);
	}

	// register terminal
	RegisterTerminal(termId, name, alias);

	//// enable terminal for comment preset +$term
	//presets[presetId].AddCommand(true, name, termId);
}

/*************************************************************************/
/* LEX::PRESET::AddCommand()                                             */
/*                                                                       */
/*************************************************************************/
void LEX::PRESET::AddCommand(bool enable, const string & name, int token /* = TOKEN_EOF */)
{
	// enable/disable command
	COMMAND comm;
	comm.enable		= enable;
	comm.value		= name;
	comm.terminal	= token;
	commands.push_back(comm);
}

/*************************************************************************/
/* LEX::Error()                                                          */
/*                                                                       */
/*************************************************************************/
void LEX::Error(void)
{
	if (Token() == TOKEN_EOF) {
		messageTable["name"] = "end of regexp";
	} else {
		messageTable["name"] = TokenValue();
	}
	
	PrintMessage(MESSAGE_LEX1006);
}

//
// End-of-file ECG_LEX_IMPL.CPP
//
