
/*************************************************************************/
/* ECG_LEX_IMPL.H                                             08.08.2006 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

#ifndef ECG_LEX_IMPL_H
#define ECG_LEX_IMPL_H

#include "ecg_lex.h"
#include "ecg_lex_lex.h"
#include "ecg_lex_msg.h"

namespace ecg {

class GENERATOR;
class LOADER_IMPL;

/*************************************************************************/
/* class LEX_IMPL                                                        */
/*                                                                       */
/*************************************************************************/
class LEX : public LEX_IMPL, public LEX_LEX {
	friend class LOADER;
public:
	typedef std::vector<int>	INT_ARRAY;
	typedef std::set<int>		INT_SET;
	typedef std::map<int, int>	INT_INT_MAP;
	typedef std::map<std::string, int> STRING_INT_MAP;
	typedef std::vector<bool>	BOOL_ARRAY;

public:
	enum PRIORITY {
		PRIORITY_NONE = -1,
		PRIORITY_LOW,
		PRIORITY_NORMAL,
		PRIORITY_HIGH,
	};

	void				Init					(int token, const std::string & regexp, PRIORITY priority, int line, std::ostream & error, bool errorOnRedefinition = true);
	void				Init					(int line, std::ostream & error);

	std::string			Line					(void) const;
	std::ostream &		GetError				(void) const;

	void				Error					(void);

	bool				GetChar					(CHAR_T & ch);

private:
	std::string			regexp;
	int					regexpPos;

	using LEX_LEX::SetPreset;

	int					token;

	int					line;
	std::ostream *		error;

public:
	typedef CHAR_INTSET			CHARS;
	typedef std::vector<CHARS>	CHARS_ARRAY;

	struct EDGE {
		CHAR_INTSET		cond;				// char-set condition (real chars that user specify)
		INT_SET			condTerm;			// condition after orthogonalization
		int				next;				// next node

		EDGE			(int next) : next(next) {}
		EDGE			(const CHAR_INTSET & cond, int next) : next(next), cond(cond) {}
	};
	typedef std::vector<EDGE> EDGE_ARRAY;

	struct NODE {
		EDGE_ARRAY		edges;
		int				token;				// TOKEN_EOF for non-final node
		PRIORITY		prior;

		NODE			(void) : token(-1), prior(PRIORITY_NONE) {}

		void			AddEdge				(const CHAR_INTSET & cond, int nextNode) { edges.push_back(EDGE(cond, nextNode)); }
		void			AddEdge				(int nextNode) { edges.push_back(EDGE(nextNode)); }
	};
	typedef std::vector<NODE> NODE_ARRAY;

	struct REGEXP {
		int				start;				// start node of regexp's graph
		int				end;				// final node of regexp's graph
	};

private:
	void				FatalError			(int messageIndex);
	void				PrintMessage		(int messageIndex);

	MESSAGE::STRING_MAP	messageTable;

private:
	struct GENTERM {
		int				id;					// generator's terminal id
		std::string		name;				// terminal's name for searching and messages
		std::string		alias;				// alias for lex generation
		bool			hasRegexp;			// mark that terminal has regexp
		int				preset;				// terminal's preset (for local terminals)

		GENTERM			(void) : id(-1), hasRegexp(false), preset(-1) {}
	};
	typedef std::vector<GENTERM> GENTERM_ARRAY;

	struct COMMAND {
		bool			enable;				// + command or - command (newValue must be empty, otherwise ignore)
		int				terminal;			// generator's terminal (TOKEN_EOF if value - nonterminal) (if newValue not empty, must be true)
		std::string		value;				// if !terminal "all" or other preset name

		COMMAND			(void) : enable(false), terminal(TOKEN_EOF) {}
	};
	typedef std::vector<COMMAND> COMMAND_ARRAY;

	struct PRESET {
		std::string		name;				// preset name
		std::string		nameUpr;			// preset name in upper case

		bool			isSkip;				// flag that skip was defined (and default value not needed)
		CHARS			skip;				// skip symbols

		COMMAND_ARRAY	commands;			// preset commands
		BOOL_ARRAY		genTerms;			// generator's terminals, true if enabled in preset
		INT_ARRAY		final;				// terminals associated with states

		enum PRODUCTIVE {
			PROD_YES,
			PROD_NO,
		};
		typedef std::vector<PRODUCTIVE> PRODUCTIVE_ARRAY;
		
		PRODUCTIVE_ARRAY productive;		// false: if from this lex's state any terminal can't be reached (for mentioned preset)

		enum STATUS {
			NOT_PROCESSED,
			PROCESSING,
			PROCESSED,
		};
		STATUS			status;

		PRESET			(void) : isSkip(false), status(NOT_PROCESSED) {}

		void			AddCommand			(bool enable, const std::string & name, int token = TOKEN_EOF);
	};
	typedef std::vector<PRESET> PRESET_ARRAY;

	struct STATE {
		INT_SET			nodes;
		INT_INT_MAP		edges;
		int				token;

		STATE			(void) : token(TOKEN_EOF) {}

		bool			operator <			(const STATE & a) const;
	};
	typedef P<STATE>	P_STATE;
	typedef std::vector<P_STATE> STATE_ARRAY;
	typedef std::map<P_STATE, int> STATE_INT_MAP;

public:
	LEX            (void);

	bool				CalcTable			(void);
	void				Generate			(const std::string & name, const std::string & nameImpl, std::ostream & header, std::ostream & code, std::ostream & hlp, GENERATOR * generator);

	void				RegisterTerminal	(int termId, const std::string & name, const std::string & alias);
	void				RegisterComment		(const std::string & preset, int termId, const std::string & name, const std::string & alias);
	bool				AddPreset			(const std::string & name, int & presetId);

private:
	int					NewNode				(void);
	bool				FixedCharSet		(char ch, CHAR_INTSET & charSet) const;
	void				Orthogonalize		(void);
	void				AddOrthogonal		(const CHARS & chars);
	void				Closing				(STATE & state);
	void				Closing				(int node, INT_INT_MAP & terminals, INT_SET & states);
	void				MakeStep			(const INT_SET & nodes, int term, INT_SET & newNodes);

	bool				StringToChars		(const char * value, int length, CHARS & ch, bool isSet = false, bool isSub = false);

	bool				CommitPresets		(void);
	bool				ProcessPreset		(PRESET & preset);
	bool				RetreiveToken		(const STATE & state, int & token, const BOOL_ARRAY & genTerms, bool skipConflicts);

	enum CGT_FLAGS {
		CGT_FILL		= 0x0001,			// fills genTerms array by local terminals
		CGT_ADD			= 0x0002,			// add default variables from existing array (can't be used with CGT_FILL)
		CGT_SUB			= 0x0004,			// subtract default variables from existing array (can't be used with CGT_FILL)
		CGT_SKIP_LOCAL	= 0x0008,			// skip local terminals (if CGT_FILL mark with false)
		CGT_ONLY_LOCAL	= 0x0010,			// only local terminals
	};
	void				CalcGenTerms		(PRESET & preset, BOOL_ARRAY & genTerms, int flags);

	typedef std::map<int, std::vector<int> > REVERSE_GRAPH;
	void				MarkProductive		(PRESET & preset, int stateId, REVERSE_GRAPH & reverseGraph);

private:
	INT_SET				startNodes;			// numbers of start nodes
	NODE_ARRAY			nodes;				// common node list
	CHARS_ARRAY			terminals;			// condition terminals
	STATE_ARRAY			table;				// parse table
	STATE_INT_MAP		searchTable;		// table to find existing states
	PRESET_ARRAY		presets;			// preset array
	STRING_INT_MAP		presetsNmb;			// table to find existing preset
	GENTERM_ARRAY		genTerms;			// generator's terminals
	INT_INT_MAP			genTermsNmb;		// generator's terms search table

	CHARS				defSkip;			// default skip symbols set
	bool				enablePresets;		// use presets when file generation
	bool				defaultWasModified; // specifies whether default preset was modified by user

	bool				errorOnRedefinition;
	int					commentLine;		// token for line comment (genTerms[genTermsNmb[commentLine]])
	int					commentBlockBegin;
	int					commentBlockEnd;

	PRIORITY			userPriority;

public:
	void				SimpleRegexp		(const std::string & ch, P<REGEXP> & res);
	void				FixedSetRegexp		(const std::string & ch, P<REGEXP> & res);
	void				SetRegexp			(P<CHARS> chars, P<REGEXP> & res);
	void				NegativeSet			(P<CHARS> a, P<CHARS> b, P<CHARS> & res);
	void				SingleChar			(const std::string & ch, P<CHARS> & res);
	void				RangeChars			(const std::string & ch, const std::string & ch2, P<CHARS> & res);
	void				FixedCharSet		(const std::string & ch, P<CHARS> & res);
	void				CharsUnion			(P<CHARS> s, P<CHARS> s2, P<CHARS> & res);
	void				ConcatRegexp		(P<REGEXP> a, P<REGEXP> b, P<REGEXP> & res);
	void				AnyRegexp			(P<REGEXP> a, P<REGEXP> & res);
	void				ManyRegexp			(P<REGEXP> a, P<REGEXP> & res);
	void				OrRegexp			(P<REGEXP> a, P<REGEXP> b, P<REGEXP> & res);
	void				YesNotRegexp		(P<REGEXP> a, P<REGEXP> & res);
	void				SetPreset			(void);
	void				DefPreset			(void);
	void				SlashPreset			(void);
	void				RestorePreset		(std::string & res);
	void				SetFinalRegexp		(P<REGEXP> a);
};

} // namespace ecg

#endif // ECG_LEX_IMPL_H
