
/*************************************************************************/
/* ECG_LOADER_IMPL.H                                          29.07.2006 */
/*                                                                       */
/* .                                                                     */
/*************************************************************************/

#ifndef ECG_LOADER_IMPL_H
#define ECG_LOADER_IMPL_H

#include "ecg_lex_impl.h"
#include "ecg_loader.h"
#include "ecg_loader_lex.h"
#include "ecg_loader_msg.h"

namespace ecg {

class GENERATOR;
struct RULE_INDEX;
struct PRIORITY;
/*************************************************************************/
/* class LOADER                                                          */
/*                                                                       */
/*************************************************************************/
class LOADER : public LOADER_IMPL, public LOADER_LEX {
	friend class GENERATOR;
public:
	typedef std::set<int> INT_SET;
	typedef std::vector<std::string> STRING_ARRAY;

public:
	void			Init				(GENERATOR & generator, std::istream & input, std::ostream & error);
	std::ostream &	GetError			(void) const { return *error; }
	void			Error				(void);
	bool			HasError			(void) const { return compilerState.value() != 0; }
	bool			GetChar				(CHAR_T & ch);

private:
	std::istream *	input;
	std::ostream *	error;
	GENERATOR *		generator;

public:
	enum ITEM_TYPE {
		TERMINAL,
		NONTERMINAL,
	};

	struct ITEM {
		ITEM_TYPE	type;
		int			id;

		ITEM		(ITEM_TYPE type, int id) : type(type), id(id) {}

		bool		operator ==		(const ITEM & item) const { return type == item.type && id == item.id; }
		bool		operator <		(const ITEM & item) const { return id < item.id || id == item.id && type < item.type; }
	};
	typedef P<ITEM>				P_ITEM;
	typedef std::vector<P_ITEM> ITEM_LIST;
	typedef std::set<P_ITEM>	ITEM_SET;

	enum RULE_ITEM_TYPE {
		SIMPLE,
		PERIODIC,
		NEM_PERIODIC,
		DEL_PERIODIC,
		YESNOT,
	};

	struct RULE_ITEM {
		P<ITEM_LIST>	itemList;
		P<ITEM_LIST>	itemList2;
		RULE_ITEM_TYPE	type;

		RULE_ITEM(P<ITEM_LIST> itemList) : type(SIMPLE), itemList(itemList), itemList2(NULL) {}
		RULE_ITEM(RULE_ITEM_TYPE type, P<ITEM_LIST> itemList) : type(type), itemList(itemList), itemList2(NULL)
		{ assert(type == PERIODIC || type == NEM_PERIODIC || type == YESNOT); }
		RULE_ITEM(P<ITEM_LIST> itemList, P<ITEM_LIST> itemList2) : type(DEL_PERIODIC), itemList(itemList), itemList2(itemList2) {}
	};
	typedef P<RULE_ITEM> P_RULE_ITEM;

	typedef std::vector<P_RULE_ITEM> RULE_OR_LIST;

	struct RULE {
		int				nontermId;
		INT_SET			productions;
	};

	struct CODE_BLOCK {
		bool			isConstructor;
		std::string		name;
		STRING_ARRAY	params;

		CODE_BLOCK		(void) : isConstructor(false) {}
		CODE_BLOCK		(const std::string & name, bool isConstructor, const STRING_ARRAY & params = STRING_ARRAY()) : name(name), params(params), isConstructor(isConstructor) {}

		bool			operator ==	(const CODE_BLOCK & cb) const;
		bool			operator != (const CODE_BLOCK & cb) const { return !(*this == cb); }
		std::string		ToString	(void) const;
	};
	typedef P<CODE_BLOCK> P_CODE_BLOCK;

	struct RULE_ATTRIB {
		bool			preferShift;
		P<ITEM_LIST>	specShift;

		bool			preferReduce;
		P<ITEM_LIST>	specReduce;

		P_CODE_BLOCK	codeBlock;
		std::string		prior;

		RULE_ATTRIB		(void) : preferShift(false), preferReduce(false) {}
	};
	typedef P<RULE_ATTRIB> P_RULE_ATTRIB;

	struct CODE_DATA {
		enum ARRAY_SUGG {
			NONE,
			INIT_ARRAY,
			APPEND_ARRAY,
		};

		P_CODE_BLOCK		code;
		int					line;
		ARRAY_SUGG			arraySugg;

		CODE_DATA			(void) : line(0), arraySugg(NONE) {}
	};
	typedef ecg::RULE_INDEX RULE_INDEX;
	typedef std::map<RULE_INDEX, CODE_DATA> RULE_CODE_MAP;
	typedef std::map<P_ITEM, CODE_DATA> ITEM_CODE_MAP;

	struct TYPE {
		std::string			name;
		bool				isValue;
		bool				isArray;
		bool				isImplMember;

		TYPE				(void) : isArray(false), isImplMember(false), isValue(false) {}
		TYPE				(const std::string & name) : name(name), isValue(true), isArray(false), isImplMember(false) {}
		TYPE				(const std::string & name, bool isImplMember, bool isValue, bool isArray) : name(name), isValue(isValue), isArray(isArray), isImplMember(isImplMember) {}

		bool				IsEmpty			(void) const { assert(!isArray || !name.empty()); return name.empty(); }
		std::string			ToString		(void) const;
		std::string			ToStrType		(void) const;
		bool				operator ==		(const std::string & name) const { if (isArray || isImplMember) return false; return this->name == name; }
		bool				operator !=		(const std::string & name) const { return !(*this == name); }
		bool				operator ==		(const TYPE & t) const { return name == t.name && isArray == t.isArray && isImplMember == t.isImplMember; }
		bool				operator !=		(const TYPE & t) const { return !(*this == t); }
	};
	typedef P<TYPE>			P_TYPE;
	typedef std::vector<P_TYPE> P_TYPE_ARRAY;

	typedef std::map<P_ITEM, P_TYPE> TYPED_ITEM_MAP;

	enum PARAM_TYPE {
		WORD,
		LITERAL,
		STRING,
	};

	struct PARAM {
		std::string			value;
		bool				fromWord;

		PARAM				(PARAM_TYPE type, const std::string & value);

		bool				IsTerminal		(void) const { assert(value.size() > 0); return !fromWord || value[0] == '$'; }
	};
	typedef P<PARAM>		P_PARAM;
	typedef std::vector<P_PARAM> PARAM_LIST;

	enum VALUE_TYPE {
		PTR,
		VAL,
	};

	typedef std::map<std::string, ecg::PRIORITY> PRIORITY_MAP;

	struct MARKER {
		int					nontermId;		// rule's nonterminal where marker was used
		int					ruleId;			// rule number where marker was used

		MARKER				(void) : nontermId(-1), ruleId(-1) {}
	};
	typedef std::map<int, MARKER> MARKER_MAP;

private:
	MESSAGE::STRING_MAP	messageTable;
	void			FatalError			(int msgId, int line = -1);
	void			PrintMessage		(int msgId, int line = -1);

private:
	int				ProcessNewRule		(int nontermId, RULE_ITEM_TYPE type, P<ITEM_LIST> ruleItem, P<ITEM_LIST> ruleItem2, INT_SET & productions);
	P_TYPE			FindNontermType		(P_ITEM nonterm, ITEM_SET & typeSearching, bool & cycleFound);
	bool			RetrieveFunctionName(P_ITEM term, std::string & code);
	bool			RetrieveFunctionName(const RULE_INDEX & ruleIndex, std::string & code);
	bool			RegisterPrefer		(P<RULE> rule, P<ITEM_LIST> itemList, bool preferShift);

    void			PrintTypedRule		(std::ostream & output, P_TYPE result, const P_TYPE_ARRAY & params, bool printEoln = true) const;
	void			GetTypesUntil		(P_TYPE_ARRAY & params, int nontermId, int ruleId, int untilNontermId = -1);

	int				AddTerminal			(P_PARAM terminal);
	std::string		GetItemName			(P_ITEM item) const;

private:
	LEX         lex;

	TYPED_ITEM_MAP	typedItems;
	RULE_CODE_MAP	ruleCode;
	ITEM_CODE_MAP	itemCode;
	int				maxPriority;
	PRIORITY_MAP	priorityAlias;

	bool			enableAutotype;		// autotype (by default it turned off)
	int				curPresetId;		// used for preset parsing
	INT_SET			needRegexp;			// set of terminals that needs regexp (like $word or $literal, but not "$word", etc.)
	MARKER_MAP		marker;				// nonterminals that used as markers

public:
	LOADER      (void)
		: maxPriority(0)
		, enableAutotype(false)
		, curPresetId(0)
	{}

public:
	void			CommitGrammar		(const std::string & name);
	void			StartNonterminal	(const std::string & name);
	void			SetAlias			(P_PARAM alias, P_PARAM param);
	void			SetItemsCodeBlock	(P<PARAM_LIST> itemList, P<CODE_BLOCK> cb);
	void			SimpleCodeBlock		(const std::string & code, P<CODE_BLOCK> & cb);
	void			CreateRule			(const std::string & nonterm, P<RULE_OR_LIST> ruleOrList, P_TYPE nontermType, P<RULE> & rule);
	void			CreateTermItem		(const std::string & term, P_ITEM & termItem);
	void			CreateWordItem		(bool onlyTerminal, const std::string & word, P_ITEM & item);
	void			CreateTermWordItem	(void);
	void			CreateBracketsItem	(P<RULE> rule, P_ITEM & item);
	void			CreateBracketsOrItem(P<RULE_OR_LIST> ruleOrList, P<TYPE> nontermType, P_ITEM & item);
	void			CreateCodeBlockItem	(P<CODE_BLOCK> cb, P_ITEM & item);
	void			ParametrizedCodeBlock(const std::string & code, P<PARAM_LIST> params, P<CODE_BLOCK> & cb);
	void			CommitRuleAndItsCodeBlock(P<RULE> rule, P<RULE_ATTRIB> prefer);
	void			CreateYesNotItem	(P<RULE_ITEM> ruleItem, P_ITEM & item);
	void			LiteralToString		(const std::string & lit, std::string & res);
	void			RegisterResultType	(P<ITEM_LIST> itemList, P_TYPE type);
	void			ProcessWordArrayType(const std::string & type, P_TYPE & res);
	void			ProcessLiteral		(std::string & res);
	void			ShiftAttrib			(P_RULE_ATTRIB attrib, P<ITEM_LIST> params, P_RULE_ATTRIB & res);
	void			ReduceAttrib		(P_RULE_ATTRIB attrib, P<ITEM_LIST> params, P_RULE_ATTRIB & res);
	void			CodeAttrib			(P_RULE_ATTRIB attrib, P_CODE_BLOCK params, P_RULE_ATTRIB & res);
	void			PriorAttrib			(P_RULE_ATTRIB attrib, const std::string prior, P_RULE_ATTRIB & res);
	void			RegisterPriority	(bool left, P<PARAM_LIST> terms);
	void			LexRegexp			(P_PARAM name, const std::string & regexp, const std::string & prior);
	void			LexSkip				(const std::string & skip);
	void			LexCLine			(const std::string & skip);
	void			LexCBegin			(const std::string & skip);
	void			LexCEnd				(const std::string & skip);
	void			LexCommand			(bool enable, P_PARAM param);
	void			LexPresetSkip		(const std::string & skip);
	void			PopPresetString		(std::string & res);
	void			CheckAliasName		(P_PARAM name, P_PARAM & res);
	void			CPT					(P_PARAM term); // check preset term
	void			LexPresetRegexp		(P_PARAM name, const std::string & regexp, const std::string & prior );
	void			LexPresetTerm		(P_PARAM term);
	void			PTNM				(const std::string & name);
	void			PTND				(const std::string & name);
	void			ESP					(void);	// enable string preset
	void			SPP					(void);
	void			PreparePreset		(void);
	void			OnAutotype			(void);
	void			EnableAutotype		(void);
	void			ENP					(void); // enable name preset
	void			ENP					(const std::string &, const std::string &, const std::string &) { ENP(); }
	void			FNT					(const std::string & name); // first non-terminal
	void			Message				(const std::string & name, const std::string & format, const std::string & info);
	void			Type				(VALUE_TYPE valueType, PARAM_TYPE paramType, const std::string & type, P_TYPE & res);
	void			TypeArray			(VALUE_TYPE valueType, PARAM_TYPE paramType, const std::string & type, P_TYPE & res);
};

} // namespace ecg

#endif // ECG_LOADER_IMPL_H
